import asyncio
import copy
import json
import logging
import random
from dataclasses import asdict
from typing import Iterable, AsyncIterable

from agere.commander import PASS_WORD, BasicJob, handler
from agere.utils.dispatcher import async_dispatcher_tools_call_for_openai
from agere.utils.llm_async_converters import LLMAsyncAdapter

from .blinker_wrapper import sync_wrapper, async_wrapper_with_loop
from .context import OpenaiContext
from .signals import (
    chat_context_extend_signal,
    common_message_signal,
    notification_signal,
    response_auxiliary_message_signal,
    response_to_user_message_stream_signal,
    response_to_user_message_sentence_stream_signal,
)
from .openai_chat_inner_service import chat_service_for_inner
from .openai_error import OpenaiErrorHandler
from .utils.openai_api import openai_api_client
from ..gptui_kernel.manager import ManagerInterface


gptui_logger = logging.getLogger("gptui_logger")


class ResponseHandler:
    """A handler to handle response from LLM."""
    def __init__(self, manager: ManagerInterface, context: OpenaiContext):
        self.manager = manager
        self.context = context

    @handler(PASS_WORD)
    async def handle_response(
        self,
        self_handler,
        response,
        at_receiving_start: list[dict] | None = None,
        at_receiving_end: list[dict] | None = None,
    ):
        """handler that handle response from LLM"""
        make_role_generator = await async_dispatcher_tools_call_for_openai(
            source=LLMAsyncAdapter(logger=gptui_logger).llm_to_async_iterable(
                response=response,
                at_receiving_start=at_receiving_start,
                at_receiving_end=at_receiving_end,
            ),
        )
        to_user_gen = make_role_generator("to_user")
        function_call_gen = make_role_generator("function_call")
        response_to_user_job = BasicJob(OpenaiHandler(self.manager, self.context).user_handler(user_gen=to_user_gen))
        await self_handler.put_job(job=response_to_user_job)
        function_call_job = BasicJob(OpenaiHandler(self.manager, self.context).function_call_handler(function_call_gen=function_call_gen))
        await self_handler.put_job(job=function_call_job)


class OpenaiHandler:
    """A handler for processing OpenAI responses"""

    def __init__(self, manager: ManagerInterface, context: OpenaiContext):
        self.manager = manager
        self.context = context
        self.chat_context_saver = context.chat_context_saver
        self.openai_api_client = openai_api_client(manager.dot_env_config_path)

    @handler(PASS_WORD)
    async def user_handler(self, self_handler, user_gen) -> None:
        """Handling the part of the message sent to the user by LLM
        args:
            user_gen: a iterable object including the message to user
        actions:
            blinker_signal:
                response_to_user_message_stream_signal: send the to_user part of chat message, which usually should be displayed.
                response_to_user_message_sentence_stream: send the voice stream divided by common punctuation marks.
        """
        empty_status = None
        collected_messages = ""
        voice_buffer = ""
        
        async for char in user_gen:
            empty_status = False
            collected_messages += char
            await response_to_user_message_stream_signal.send_async(
                self,
                _sync_wrapper=sync_wrapper,
                message={"content": {"content": char, "context_id": self.context.id}, "flag": "content"},
            )
            
            # Send voice signal
            if response_to_user_message_sentence_stream_signal.receivers:
                voice_buffer += char
                if char.startswith((".","!","?",";",":","。","！","？","；","：","\n")):
                    await response_to_user_message_sentence_stream_signal.send_async(
                        self,
                        _sync_wrapper=sync_wrapper,
                        message={"content": voice_buffer, "flag": "content"},
                    )
                    voice_buffer = ""
        
        if empty_status is None:
            return
        if voice_buffer:
            await response_to_user_message_sentence_stream_signal.send_async(
                self,
                _sync_wrapper=sync_wrapper,
                message={"content": voice_buffer, "flag": "content"},
            )
            voice_buffer = ""

        await response_to_user_message_stream_signal.send_async(
            self,
            _sync_wrapper=sync_wrapper,
            message={"content": {"content": "", "context_id": self.context.id}, "flag": "end"},
        )
        
        # save response to context
        if self.chat_context_saver == "outer":
            await chat_context_extend_signal.send_async(
                self,
                _sync_wrapper=sync_wrapper,
                message={
                    "content": {
                        "messages": [{"role": "assistant", "content": collected_messages}],
                        "context": self.context,
                    },
                    "flag": "",
                }
            )
        else:
            self.context.chat_context_append({"role": "assistant", "content": collected_messages})
        # Send voice end signal
        if response_to_user_message_sentence_stream_signal.receivers:
            await response_to_user_message_sentence_stream_signal.send_async(self, _sync_wrapper=sync_wrapper, message={"content":"", "flag":"end"})
    
    @handler(PASS_WORD)
    async def function_call_handler(self, self_handler, function_call_gen) -> None:
        function_result_dict = {}
        async for function_call in function_call_gen:
            if not function_call:
                continue
            function_dict = {}
            try:
                function_dict = json.loads(function_call)
            except json.JSONDecodeError as e:
                gptui_logger.error(f"An error occurred while parsing the JSON string. JSON string: {function_call}. Error: {e}")
            if function_dict.get("name"):
                # call the function
                tool_call_index = function_dict["tool_call_index"]
                tool_call_id = function_dict["tool_call_id"]
                function_name = function_dict["name"]
                function_to_call = self.manager.available_functions_link[function_name]
                function_args = function_dict["arguments"]
                
                await notification_signal.send_async(
                    self,
                    _sync_wrapper=sync_wrapper,
                    message={
                        "content":{
                            "content": f"Function calling{f'({tool_call_index})' if tool_call_index else ''}: " + function_name + " ... ",
                            "description": "raw",
                        },
                        "flag": "info",
                    }
                )

                context = self.manager.gk_kernel.context_render(args=function_args, function=function_to_call)
                
                # Dose insert context
                if context.variables.get("openai_context") == "AUTO":
                    openai_context_deepcopy = copy.deepcopy(self.context)
                    openai_context_deepcopy.plugins = [repr(plugin) for plugin in openai_context_deepcopy.plugins]
                    context["openai_context"] = json.dumps(asdict(openai_context_deepcopy))

                function_call_display_str = f"{function_name}({', '.join(f'{k}={v}' for k, v in function_args.items())})"
                await response_auxiliary_message_signal.send_async(
                    self,
                    _sync_wrapper=sync_wrapper,
                    message={
                        "content":{"role": "gpt", "content": f"Function call: {function_call_display_str}"},
                        "flag":"function_call",
                    }
                )
                gptui_logger.info(f"Function call: {function_call_display_str}; ID: {tool_call_id}.")
                function_response_context = await function_to_call.invoke_async(context=context)
                function_response = str(function_response_context)
                function_result_dict[tool_call_index] = {
                    "tool_call_id": tool_call_id,
                    "function_name": function_name,
                    "function_args": function_args,
                    "function_result": function_response,
                }
                await response_auxiliary_message_signal.send_async(
                    self,
                    _sync_wrapper=sync_wrapper,
                    message={
                        "content":
                            {
                                "role": "function",
                                "name": function_name,
                                "content": function_response,
                            },
                        "flag":"function_response",
                    },
                )
        
        if not function_result_dict:
            return
        # send the function response to GPT
        
        message = [
            # tools call info
            {
                "role": "assistant",
                "content": None,
                "tool_calls": [
                    {
                        "id": one_function_call["tool_call_id"],
                        "function": {
                            "arguments": str(one_function_call["function_args"]),
                            "name": one_function_call["function_name"]
                        },
                        "type": "function"
                    } for one_function_call in function_result_dict.values()
                ]
            }
        ] + \
        [
            # tools call results
            {
                "tool_call_id": function_result["tool_call_id"],
                "role": "tool",
                "name": function_result["function_name"],
                "content": function_result["function_result"],
            } for function_result in function_result_dict.values()
        ]

        gptui_logger.info(f"Function response: {message}")
        functions = self.manager.available_functions_meta
        await notification_signal.send_async(
            self,
            _sync_wrapper=sync_wrapper,
            message={
                "content": {
                    "content": "Sent function call result, waiting for response ...",
                    "description": "raw",
                },
                "flag": "info",
            }
        )
        if functions:
            paras = {"messages_list": message, "context": self.context, "openai_api_client": self.openai_api_client, "tools": functions, "tool_choice": "auto"}
        else:
            paras = {"messages_list": message, "context": self.context, "openai_api_client": self.openai_api_client}
        
        # retry twice
        last_exception = None
        for _ in range(2):
            try:
                response = await asyncio.to_thread(chat_service_for_inner, **paras)
                break
            except Exception as e:
                last_exception = e
                await asyncio.sleep(5)
        else:
            assert last_exception is not None
            OpenaiErrorHandler().openai_error_handle(error=last_exception, context=self.context)
            self.context.chat_context_append(
                {
                    "role": "user",
                    "content": (
                        f"You attempted to call the following tools: {str([result['function_name'] for result in function_result_dict.values()])}, "
                        "but encountered an exception. If necessary, you can retry or try again later."
                    )
                }
            )
            raise last_exception

        if self.chat_context_saver == "outer":
            await chat_context_extend_signal.send_async(
                self,
                _sync_wrapper=sync_wrapper,
                message={
                    "content": {
                        "messages": message,
                        "context": self.context
                    },
                    "flag":"function_response"}
            )
        else:
            for one_message in message:
                self.context.chat_context_append(one_message)

        ResponseJob = self.manager.get_job("ResponseJob")
        at_receiving_start = [
            {
                "function": notification_signal.send,
                "params": {
                    "args": (self,),
                    "kwargs": {
                        "_async_wrapper": async_wrapper_with_loop,
                        "message":{
                            "content":{
                                "content": "Starting to receive response messages for function calls.",
                                "description": "raw",
                            },
                            "flag":"info",
                        },
                    },
                },
            },
        ]
        await self_handler.put_job(
            ResponseJob(
                response=response,
                manager=self.manager,
                context=self.context,
                at_receiving_start=at_receiving_start,
            )
        )


class GroupTalkHandler:
    @handler(PASS_WORD)
    async def handle_response(self, self_handler, response_dict: dict[str, Iterable]):
        items = list(response_dict.items())
        # Randomly shuffle the order to give each role an equal opportunity to speak.
        random.shuffle(items)
        for role_name, response in items:
            self_handler.call_handler(self.parse_stream_response(role_name=role_name, stream_response=response))

    @handler(PASS_WORD)
    async def wait_for_termination(self, self_handler, group_talk_manager):
        while True:
            await asyncio.sleep(0.5)
            # If no one is speaking, check if the user is speaking.
            if group_talk_manager.speaking is None:
                messages = [{"role": "user", "name": group_talk_manager.user_name, "content": message} for message in group_talk_manager.user_talk_buffer]
                if messages:
                    with group_talk_manager._ai_care_rlock:
                        group_talk_manager.ai_care_message_buffer = []
                    common_message_signal.send(
                        self,
                        _async_wrapper=async_wrapper_with_loop,
                        message={
                            "content": messages,
                            "flag": "group_talk_user_message_send",
                        },
                    )
                    response_dict = {}
                    items = list(group_talk_manager.roles.items())
                    # Randomly shuffle the order to give each role an equal opportunity to speak.
                    random.shuffle(items)
                    for role_name, role in items:
                        response_dict[role_name] = role.chat(message=messages)
                    GroupTalkHandler = group_talk_manager.manager.get_handler("GroupTalkHandler")
                    # Because at any given moment only one role can speak, it is safe when multiple dialogue tasks are running in parallel;
                    # there will be no inconsistency in the role's chat history.
                    self_handler.call_handler(GroupTalkHandler().handle_response(response_dict))
                    group_talk_manager.user_talk_buffer = []
                    for role in group_talk_manager.roles.values():
                        role.ai_care_update()
                else:
                    # The task might iteratively receive messages and incur certain delays, but that's okay; 
                    # there is a character currently speaking, and it is permissible for the loop to be blocked at this time.
                    with group_talk_manager._ai_care_rlock:
                        await group_talk_manager.send_ai_care_message()
            if not group_talk_manager.running:
                await group_talk_manager.close_task_node()
                group_talk_manager.speaking = None
                break

    @handler(PASS_WORD)
    async def parse_stream_response(self, self_handler, role_name, stream_response):
        async_stream_response = LLMAsyncAdapter().llm_to_async_iterable(stream_response)
        talk_manager = self_handler.ancestor_chain[-2]
        if not talk_manager.running:
            return
        full_response_content = await self.stream_response_result(async_stream_response=async_stream_response)
        role = talk_manager.roles[role_name]
        if "Can I speak" in full_response_content:
            if talk_manager.speaking:
                # No need to actually reply.
                role.context.chat_context_append({"role": "assistant", "content": "Can I speak?"})
                role.context.chat_context_append({"role": "user", "name": "host", "content": f"SYS_INNER: Host says to you: No, {talk_manager.speaking} is speaking."})
                gptui_logger.info(f"{role_name} want to speak, but {talk_manager.speaking} is speaking.")
            else:
                gptui_logger.info(f"{role_name} speak.")
                role.context.chat_context_append({"role": "assistant", "content": "Can I speak?"})
                talk_manager.speaking = role_name
                try:
                    response = talk_manager.roles[role_name].chat(
                        message={
                            "role": "user",
                            "name": "host",
                            "content": f"SYS_INNER: Host says to you: Yes, you are {role_name}, you can talk now. Reply directly with what you want to say, without additionally thanking the host.",
                        }
                    )
                    async_talk_stream_response = LLMAsyncAdapter().llm_to_async_iterable(response)
                    talk_content = await self.stream_response_display_and_result(role_name=role_name, async_stream_response=async_talk_stream_response, talk_manager=talk_manager)
                except Exception as e:
                    # If receiving the message fails, set talk_manager.speaking to None to avoid blocking the group chat.
                    await talk_manager.set_speaking_to_none()
                    gptui_logger.warning(f"Encountered an error when receiving the speech content of group chat member {role_name}, error: {e}")
                else:
                    TalkToAll = talk_manager.manager.get_job("TalkToAll")
                    await self_handler.put_job(TalkToAll(message_content=talk_content, message_from=role_name))
        else:
            role.context.chat_context_append({"role": "assistant", "content": " "})
            if full_response_content and not full_response_content.isspace():
                gptui_logger.info(f"The group talk member {role_name} speak without asking 'Can I speak?' first, full_response_content: {full_response_content}")
            else:
                gptui_logger.info(f"{role_name} stay silent.")

    async def stream_response_result(self, async_stream_response: AsyncIterable) -> str:
        chunk_list = []
        async for chunk in async_stream_response:
            chunk_content = chunk.choices[0].delta.content
            if not chunk_content:
                continue
            chunk_list.append(chunk_content)
        full_response_content = ''.join(chunk_list)
        return full_response_content      

    async def stream_response_display_and_result(self, role_name: str, async_stream_response: AsyncIterable, talk_manager) -> str:
        chunk_list = []
        async for chunk in async_stream_response:
            chunk_content = chunk.choices[0].delta.content
            if not chunk_content:
                continue
            await response_auxiliary_message_signal.send_async(
                self,
                _sync_wrapper=sync_wrapper,
                message={
                    "content": {
                        "flag": "content",
                        "content": {"role": "assistant", "name": role_name, "content": chunk_content, "group_talk_manager_id": talk_manager.group_talk_manager_id},
                    },
                    "flag": "group_talk_response",
                },
            )
            chunk_list.append(chunk_content)
        full_response_content = ''.join(chunk_list)
        await response_auxiliary_message_signal.send_async(
            self,
            _sync_wrapper=sync_wrapper,
            message={
                "content": {
                    "flag": "end",
                    "content": {"role": "assistant", "name": role_name, "content": "", "group_talk_manager_id": talk_manager.group_talk_manager_id},
                },
                "flag": "group_talk_response",
            },
        )
        return full_response_content



import logging
import sys
import time
import traceback
import types

import litellm
import tokentrim as tt
from rich import print
from rich.markdown import Markdown


class FakeReadlineContext:
    """
    When used in wezterm, invoking open-interpreter with its import readline causes the TUI to freeze and display anomalies.
    Since readline in open-interpreter is not useful for this project,
    it is temporarily replaced here to prevent it from importing readline.
    """
    
    def __enter__(self):
        self.original_readline = sys.modules.get("readline")
        fake_readline_module = types.ModuleType("f_readline")
        def add_history(*args, **kwargs):
            """Only this function was used in open-interpreter, so only a replacement for this function is provided."""
            pass
        fake_readline_module.add_history = add_history
        sys.modules["readline"] = fake_readline_module
    
    def __exit__(self, exc_type, exc_value, exc_tb):
        if self.original_readline:
            sys.modules["readline"] = self.original_readline
        else:
            del sys.modules["readline"]


with FakeReadlineContext():
    import readline
    from interpreter.code_block import CodeBlock
    from interpreter.code_interpreter import CodeInterpreter
    from interpreter.get_hf_llm import get_hf_llm
    from interpreter.interpreter import Interpreter, function_schema, confirm_mode_message
    from interpreter.message_block import MessageBlock
    from interpreter.utils import merge_deltas, parse_partial_json


gptui_logger = logging.getLogger("gptui_logger")


class MyInterpreter(Interpreter):
    
    # Override the 'chat' method, replacing call respond to yield from respond.
    def chat(self, message=None, return_messages=False):
        # Connect to an LLM (an large language model)
        if not self.local:
            # gpt-4
            self.verify_api_key()

        # ^ verify_api_key may set self.local to True, so we run this as an 'if', not 'elif':
        if self.local:

            # Code-Llama
            if self.llama_instance == None:

                # Find or install Code-Llama
                try:
                    self.llama_instance = get_hf_llm(self.model, self.debug_mode, self.context_window)
                    if self.llama_instance == None:
                        # They cancelled.
                        return
                except:
                    traceback.print_exc()
                    # If it didn't work, apologize and switch to GPT-4

                    print(Markdown("".join([
                        f"> Failed to install `{self.model}`.",
                        f"\n\n**Common Fixes:** You can follow our simple setup docs at the link below to resolve common errors.\n\n```\nhttps://github.com/KillianLucas/open-interpreter/tree/main/docs\n```",
                        f"\n\n**If you've tried that and you're still getting an error, we have likely not built the proper `{self.model}` support for your system.**",
                        "\n\n*( Running language models locally is a difficult task!* If you have insight into the best way to implement this across platforms/architectures, please join the Open Interpreter community Discord and consider contributing the project's development. )",
                        "\n\nPress enter to switch to `GPT-4` (recommended)."
                    ])))
                    input()

                    # Switch to GPT-4
                    self.local = False
                    self.model = "gpt-4"
                    self.verify_api_key()

        # Display welcome message
        welcome_message = ""

        if self.debug_mode:
            welcome_message += "> Entered debug mode"

        # If self.local, we actually don't use self.model
        # (self.auto_run is like advanced usage, we display no messages)
        if not self.local and not self.auto_run:

            if self.use_azure:
                notice_model = f"{self.azure_deployment_name} (Azure)"
            else:
                notice_model = f"{self.model.upper()}"
            welcome_message += f"\n> Model set to `{notice_model}`\n\n**Tip:** To run locally, use `interpreter --local`"

        if self.local:
            welcome_message += f"\n> Model set to `{self.model}`"

        # If not auto_run, tell the user we'll ask permission to run code
        # We also tell them here how to exit Open Interpreter
        if not self.auto_run:
            welcome_message += "\n\n" + confirm_mode_message

        welcome_message = welcome_message.strip()


        # Modified by happyapplehorse ##########################################
        # Original:
        # Print welcome message with newlines on either side (aesthetic choice)
        # unless we're starting with a blockquote (aesthetic choice)
        #if welcome_message != "":
        #    if welcome_message.startswith(">"):
        #        print(Markdown(welcome_message), '')
        #    else:
        #        print('', Markdown(welcome_message), '')
        # Changed to:

        ########################################################################


        # Check if `message` was passed in by user
        if message:
            # If it was, we respond non-interactivley
            self.messages.append({"role": "user", "content": message})
            # Modified by happyapplehorse ##########################################
            # Original:
            #self.respond()
            # Changed to:
            out = yield from self.respond()
            return out
            ########################################################################

        else:
            # If it wasn't, we start an interactive chat
            while True:
                try:
                    user_input = input("> ").strip()
                except EOFError:
                    break
                except KeyboardInterrupt:
                    print()  # Aesthetic choice
                    break

                # Use `readline` to let users up-arrow to previous user messages,
                # which is a common behavior in terminals.
                readline.add_history(user_input)

                # If the user input starts with a `%` or `/`, it's a command
                if user_input.startswith("%") or user_input.startswith("/"):
                    self.handle_command(user_input)
                    continue

                # Add the user message to self.messages
                self.messages.append({"role": "user", "content": user_input})

                # Respond, but gracefully handle CTRL-C / KeyboardInterrupt
                try:
                    self.respond()
                except KeyboardInterrupt:
                    pass
                finally:
                    # Always end the active block. Multiple Live displays = issues
                    self.end_active_block()

        if return_messages:
            return self.messages
    
    # Override the 'respond' method, replacing interactions with the terminal with 'yield'.
    def respond(self):
        # Add relevant info to system_message
        # (e.g. current working directory, username, os, etc.)
        info = self.get_info_for_system_message()

        # This is hacky, as we should have a different (minified) prompt for CodeLLama,
        # but for now, to make the prompt shorter and remove "run_code" references, just get the first 2 lines:
        if self.local:
            self.system_message = "\n".join(self.system_message.split("\n")[:2])
            self.system_message += "\nOnly do what the user asks you to do, then ask what they'd like to do next."

        system_message = self.system_message + "\n\n" + info

        if self.local:
            messages = tt.trim(self.messages, max_tokens=(self.context_window - self.max_tokens - 25),
                               system_message=system_message)
        else:
            messages = tt.trim(self.messages, self.model, system_message=system_message)

        if self.debug_mode:
            print("\n", "Sending `messages` to LLM:", "\n")
            print(messages)
            print()

        # Make LLM call
        if not self.local:
            # GPT

            error = ""

            for _ in range(3):  # 3 retries
                try:

                    if self.use_azure:
                        response = litellm.completion(
                            f"azure/{self.azure_deployment_name}",
                            messages=messages,
                            functions=[function_schema],
                            temperature=self.temperature,
                            stream=True,
                        )
                    else:
                        if self.api_base:
                            # The user set the api_base. litellm needs this to be "custom/{model}"
                            response = litellm.completion(
                                api_base=self.api_base,
                                model="custom/" + self.model,
                                messages=messages,
                                functions=[function_schema],
                                stream=True,
                                temperature=self.temperature,
                            )
                        else:
                            # Normal OpenAI call
                            response = litellm.completion(
                                model=self.model,
                                messages=messages,
                                functions=[function_schema],
                                stream=True,
                                temperature=self.temperature,
                            )

                    break
                except:
                    if self.debug_mode:
                        traceback.print_exc()
                    error = traceback.format_exc()
                    time.sleep(3)
            else:
                raise Exception(error)

        elif self.local:
            # Code-Llama

            # Convert messages to prompt
            # (This only works if the first message is the only system message)

            def messages_to_prompt(messages):

                for message in messages:
                    # Happens if it immediatly writes code
                    if "role" not in message:
                        message["role"] = "assistant"

                # Falcon prompt template
                if "falcon" in self.model.lower():

                    formatted_messages = ""
                    for message in messages:
                        formatted_messages += f"{message['role'].capitalize()}: {message['content']}\n"
                    formatted_messages = formatted_messages.strip()

                else:
                    # Llama prompt template

                    # Extracting the system prompt and initializing the formatted string with it.
                    system_prompt = messages[0]['content']
                    formatted_messages = f"<s>[INST] <<SYS>>\n{system_prompt}\n<</SYS>>\n"

                    # Loop starting from the first user message
                    for index, item in enumerate(messages[1:]):
                        role = item['role']
                        content = item['content']

                        if role == 'user':
                            formatted_messages += f"{content} [/INST] "
                        elif role == 'function':
                            formatted_messages += f"Output: {content} [/INST] "
                        elif role == 'assistant':
                            formatted_messages += f"{content} </s><s>[INST] "

                    # Remove the trailing '<s>[INST] ' from the final output
                    if formatted_messages.endswith("<s>[INST] "):
                        formatted_messages = formatted_messages[:-10]

                return formatted_messages

            prompt = messages_to_prompt(messages)
            # Lmao i can't believe this works (it does need this btw)
            if messages[-1]["role"] != "function":
                prompt += "Let's explore this. By the way, I can run code on your machine by writing the code in a markdown code block. This works for shell, javascript, python, R, and applescript. I'm going to try to do this for your task. Anyway, "
            elif messages[-1]["role"] == "function" and messages[-1]["content"] != "No output":
                prompt += "Given the output of the code I just ran, "
            elif messages[-1]["role"] == "function" and messages[-1]["content"] == "No output":
                prompt += "Given the fact that the code I just ran produced no output, "

            if self.debug_mode:
                # we have to use builtins bizarrely! because rich.print interprets "[INST]" as something meaningful
                import builtins
                builtins.print("TEXT PROMPT SEND TO LLM:\n", prompt)

            # Run Code-Llama

            response = self.llama_instance(
                prompt,
                stream=True,
                temperature=self.temperature,
                stop=["</s>"],
                max_tokens=750  # context window is set to 1800, messages are trimmed to 1000... 700 seems nice
            )

        # Initialize message, function call trackers, and active block
        self.messages.append({})
        in_function_call = False
        llama_function_call_finished = False
        self.active_block = None

        for chunk in response:
            if self.use_azure and ('choices' not in chunk or len(chunk['choices']) == 0):
                # Azure OpenAI Service may return empty chunk
                continue

            if self.local:
                if "content" not in messages[-1]:
                    # This is the first chunk. We'll need to capitalize it, because our prompt ends in a ", "
                    chunk["choices"][0]["text"] = chunk["choices"][0]["text"].capitalize()
                    # We'll also need to add "role: assistant", CodeLlama will not generate this
                    messages[-1]["role"] = "assistant"
                delta = {"content": chunk["choices"][0]["text"]}
            else:
                delta = chunk["choices"][0]["delta"]

            # Accumulate deltas into the last message in messages
            self.messages[-1] = merge_deltas(self.messages[-1], delta)

            # Check if we're in a function call
            if not self.local:
                condition = "function_call" in self.messages[-1]
            elif self.local:
                # Since Code-Llama can't call functions, we just check if we're in a code block.
                # This simply returns true if the number of "```" in the message is odd.
                if "content" in self.messages[-1]:
                    condition = self.messages[-1]["content"].count("```") % 2 == 1
                else:
                    # If it hasn't made "content" yet, we're certainly not in a function call.
                    condition = False
            
            if condition:
                # We are in a function call.

                # Check if we just entered a function call
                if in_function_call == False:

                    # If so, end the last block,
                    self.end_active_block()

                    # Print newline if it was just a code block or user message
                    # (this just looks nice)
                    last_role = self.messages[-2]["role"]
                    if last_role == "user" or last_role == "function":
                        # Modified by happyapplehorse ##########################################
                        # Original:
                        #print()
                        # Changed to:
                        pass
                        ########################################################################

                    # then create a new code block
                    # Modified by happyapplehorse ##########################################
                    # Original:
                    #self.active_block = CodeBlock()
                    # Changed to:
                    self.active_block = MyCodeBlock()
                    ########################################################################

                # Remember we're in a function_call
                in_function_call = True

                # Now let's parse the function's arguments:

                if not self.local:
                    # gpt-4
                    # Parse arguments and save to parsed_arguments, under function_call
                    if "arguments" in self.messages[-1]["function_call"]:
                        arguments = self.messages[-1]["function_call"]["arguments"]
                        new_parsed_arguments = parse_partial_json(arguments)
                        if new_parsed_arguments:
                            # Only overwrite what we have if it's not None (which means it failed to parse)
                            self.messages[-1]["function_call"][
                                "parsed_arguments"] = new_parsed_arguments

                elif self.local:
                    # Code-Llama
                    # Parse current code block and save to parsed_arguments, under function_call
                    if "content" in self.messages[-1]:

                        content = self.messages[-1]["content"]

                        if "```" in content:
                            # Split by "```" to get the last open code block
                            blocks = content.split("```")

                            current_code_block = blocks[-1]

                            lines = current_code_block.split("\n")

                            if content.strip() == "```":  # Hasn't outputted a language yet
                                language = None
                            else:
                                if lines[0] != "":
                                    language = lines[0].strip()
                                else:
                                    language = "python"
                                    # In anticipation of its dumbassery let's check if "pip" is in there
                                    if len(lines) > 1:
                                        if lines[1].startswith("pip"):
                                            language = "shell"

                            # Join all lines except for the language line
                            code = '\n'.join(lines[1:]).strip("` \n")

                            arguments = {"code": code}
                            if language:  # We only add this if we have it-- the second we have it, an interpreter gets fired up (I think? maybe I'm wrong)
                                if language == "bash":
                                    language = "shell"
                                arguments["language"] = language

                        # Code-Llama won't make a "function_call" property for us to store this under, so:
                        if "function_call" not in self.messages[-1]:
                            self.messages[-1]["function_call"] = {}

                        self.messages[-1]["function_call"]["parsed_arguments"] = arguments

            else:
                # We are not in a function call.

                # Check if we just left a function call
                if in_function_call == True:

                    if self.local:
                        # This is the same as when gpt-4 gives finish_reason as function_call.
                        # We have just finished a code block, so now we should run it.
                        llama_function_call_finished = True

                # Remember we're not in a function_call
                in_function_call = False

                # If there's no active block,
                if self.active_block == None:
                    # Create a message block
                    # Modified by happyapplehorse ##########################################
                    # Original:
                    #self.active_block = MessageBlock()
                    # Changed to:
                    self.active_block = MyMessageBlock()
                    ########################################################################

            # Update active_block
            self.active_block.update_from_message(self.messages[-1])

            # Check if we're finished
            if chunk["choices"][0]["finish_reason"] or llama_function_call_finished:
                if chunk["choices"][
                    0]["finish_reason"] == "function_call" or llama_function_call_finished:
                    # Time to call the function!
                    # (Because this is Open Interpreter, we only have one function.)

                    if self.debug_mode:
                        print("Running function:")
                        print(self.messages[-1])
                        print("---")

                    # Ask for user confirmation to run code
                    if self.auto_run == False:

                        # End the active block so you can run input() below it
                        # Save language and code so we can create a new block in a moment
                        self.active_block.end()
                        language = self.active_block.language
                        code = self.active_block.code

                        # Modified by happyapplehorse ##########################################
                        
                        # Original:
                        # Prompt user
                        #response = input("  Would you like to run this code? (y/n)\n\n  ")
                        #print("")  # <- Aesthetic choice

                        # Changed to:
                        response = yield self.messages[-1]

                        ########################################################################

                        if response.strip().lower() == "y":
                            # Create a new, identical block where the code will actually be run
                            # Modified by happyapplehorse ##########################################
                            # Original:
                            #self.active_block = CodeBlock()
                            # Changed to:
                            self.active_block = MyCodeBlock()
                            ########################################################################
                            self.active_block.language = language
                            self.active_block.code = code

                        else:
                            # User declined to run code.
                            self.active_block.end()
                            self.messages.append({
                                "role":
                                    "function",
                                "name":
                                    "run_code",
                                "content":
                                    "User decided not to run this code."
                            })
                            return self.messages[-1]

                    # If we couldn't parse its arguments, we need to try again.
                    if not self.local and "parsed_arguments" not in self.messages[-1]["function_call"]:
                        # After collecting some data via the below instruction to users,
                        # This is the most common failure pattern: https://github.com/KillianLucas/open-interpreter/issues/41

                        # print("> Function call could not be parsed.\n\nPlease open an issue on Github (openinterpreter.com, click Github) and paste the following:")
                        # print("\n", self.messages[-1]["function_call"], "\n")
                        # time.sleep(2)
                        # print("Informing the language model and continuing...")

                        # Since it can't really be fixed without something complex,
                        # let's just berate the LLM then go around again.

                        self.messages.append({
                            "role": "function",
                            "name": "run_code",
                            "content": """Your function call could not be parsed. Please use ONLY the `run_code` function, which takes two parameters: `code` and `language`. Your response should be formatted as a JSON."""
                        })

                        self.respond()
                        return self.messages[-1]

                    # Create or retrieve a Code Interpreter for this language
                    language = self.messages[-1]["function_call"]["parsed_arguments"][
                        "language"]
                    if language not in self.code_interpreters:
                        self.code_interpreters[language] = CodeInterpreter(language, self.debug_mode)
                    code_interpreter = self.code_interpreters[language]

                    # Let this Code Interpreter control the active_block
                    code_interpreter.active_block = self.active_block
                    code_interpreter.run()

                    # End the active_block
                    self.active_block.end()

                    # Append the output to messages
                    # Explicitly tell it if there was no output (sometimes "" = hallucinates output)
                    self.messages.append({
                        "role": "function",
                        "name": "run_code",
                        "content": self.active_block.output if self.active_block.output else "No output"
                    })

                    # Go around again
                    # Modified by happyapplehorse ##########################################
                    # Original:
                    #self.respond()
                    # Changed to:
                    out = yield from self.respond()
                    return out
                    ########################################################################

                if chunk["choices"][0]["finish_reason"] != "function_call":
                    # Done!

                    # Code Llama likes to output "###" at the end of every message for some reason
                    if self.local and "content" in self.messages[-1]:
                        self.messages[-1]["content"] = self.messages[-1]["content"].strip().rstrip("#")
                        self.active_block.update_from_message(self.messages[-1])
                        time.sleep(0.1)

                    self.active_block.end()
                    print("5555")
                    print(self.messages)
                    return self.messages[-1]

        # Modified by happyapplehorse ##########################################
        # Original:

        # Changed to:
        return self.messages[-1]
        ########################################################################


class MyMessageBlock(MessageBlock):
    def update_from_message(self, message: dict) -> None:
        pass


class MyCodeBlock(CodeBlock):
    def update_from_message(self, message: dict) -> None:
        if "function_call" in message and "parsed_arguments" in message["function_call"]:

            parsed_arguments = message["function_call"]["parsed_arguments"]

            if parsed_arguments != None:
                self.language = parsed_arguments.get("language")
                self.code = parsed_arguments.get("code")


def response_render(message: dict):
    if message is None:
        return "null"
    out = message.get("content") or ""
    out += "\n\n"
    command = message.get("function_call", {}).get("parsed_arguments", None)
    code = language = None
    if command is not None:
        language = command.get("language")
        code = command.get("code")
    if language and code:
        out += (
            f"I will execute \"{code}\" in \"{language}\""
            "\n\nWould you like to run this code? (y/n)\n"
            "You should inform the user about the code I intend to run, "
            "ask for his/her consent, and then precisely reply to me with either 'y' or 'n'."
        )
    return out




import logging

from .blinker_wrapper import async_wrapper_with_loop, async_wrapper_without_loop
from .context import OpenaiContext
from .signals import notification_signal


gptui_logger = logging.getLogger("gptui_logger")


class OpenaiErrorHandler:
    def openai_error_handle(self, error: Exception, context: OpenaiContext, event_loop: bool = True, **kwargs) -> None:
        if event_loop is True:
            gptui_logger.error(f"Openai Error: {error}")
            notification_signal.send(self, _async_wrapper=async_wrapper_with_loop, message={"content":{"error":error, "context":context, "ps":kwargs}, "flag":"openai_error"})
        else:
            gptui_logger.error(f"Openai Error: {error}")
            notification_signal.send(self, _async_wrapper=async_wrapper_without_loop, message={"content":{"error":error, "context":context, "ps":kwargs}, "flag":"openai_error"})



from __future__ import annotations
from typing import Callable, TYPE_CHECKING


if TYPE_CHECKING:
    from .driver_interface import DriverInterface


class DriverError(Exception):
    ...


class NoDriverError(DriverError):
    def __init__(self, driver: str):
        self.driver = driver
    
    def __str__(self):
        return f"There is no {self.driver} driver."


class NoDriverMethodError(DriverError):
    def __init__(self, driver: str | Callable, method: str | DriverInterface):
        if isinstance(driver, str):
            self.driver = driver
        else:
            self.driver = driver.__name__
        if isinstance(method, str):
            self.method = method
        else:
            self.method = type(method).__name__
    
    def __str__(self):
        return f"There is no {self.method} method in {self.driver} driver."



import json
import logging

from semantic_kernel.orchestration.sk_context import SKContext
from semantic_kernel.skill_definition import sk_function, sk_function_context_parameter

from gptui.gptui_kernel.manager import auto_init_params


gptui_logger = logging.getLogger("gptui_logger")


class Memo:
    def __init__(self, app):
        self.app = app

    @auto_init_params("0")
    @classmethod
    def get_init_params(cls, manager) -> tuple:
        return (manager.client,)
    
    @sk_function(
        description="Record important information; the content should be significant and concise.",
        name="write_memo",
    )
    @sk_function_context_parameter(
        name="content",
        description="Information to be written into the memo.",
    )
    @sk_function_context_parameter(
        name="openai_context",
        description=(
            "The dictionary string version of the OpenaiContext instance. "
            "This is a special parameter that typically doesn't require manual intervention, as it is usually automatically managed."
            "Unless there's a clear intention, please keep its default value."
        ),
        default_value="AUTO"
    )
    def write_memo(self, context: SKContext) -> str:
        content = context["content"]
        try:
            openai_context_dict = json.loads(str(context["openai_context"]))
        except json.JSONDecodeError:
            return ("An error occurred while parsing the openai_context content. "
                "You should not provide the 'openai_context' parameter as the system automatically supplies it."
            )
        conversation_id = int(openai_context_dict["id"])
        try:
            conversation = self.app.openai.conversation_dict[conversation_id]
        except KeyError:
            return f"Write memo faild. Conversation id {conversation_id} is not correct."
        openai_context = conversation["openai_context"]
        bead_content = openai_context.bead
        bead_content[-1]["content"] += "\n" + content
        openai_context.insert_bead()
        return f"'{content}' have been written into the memo!" 



import os

from ..data.langchain.document_loaders import TextLoader
from ..data.langchain.schema import Document


class Doc:
    def __init__(self, doc_name: str, doc_ext: str, pointer, description: str | None = None):
        self.name = doc_name
        self.ext = doc_ext
        self.pointer = pointer
        self.description = description
        if isinstance(pointer, Document):
            self.content_type = "Document"
        elif isinstance(pointer, str):
            self.content_type = "str"
        else:
            self.content_type = "Unknown"

    @property
    def content(self):
        if isinstance(self.pointer, Document):
            return self.pointer.page_content
        else:
            return self.pointer
    
def document_loader(file_path: str) -> list[Document]:
    file_ext_name = os.path.splitext(file_path)[1]
    if file_ext_name in {".txt", ".md", ".json", ".py", ".cpp", ".yaml", ".yml", ".toml", ".log"}:
        loader = TextLoader(file_path)
    else:
        raise TypeError("Selected file type is not suppported.")
    document_list = loader.load()
    return document_list



import logging

from .dash_board_control import DashBoard
from ..models.signals import chat_context_extend_signal, chat_context_extend_for_sending_signal
from ..views.common_message import CommonMessage

gptui_logger = logging.getLogger("gptui_logger")


class ChatContextControl:
    def __init__(self, app):
        self.app = app
        self.dash_board = DashBoard(app)
        self.chat_context_to_vectorize_buffer = {}
        chat_context_extend_signal.connect(self.chat_context_extend)
        chat_context_extend_for_sending_signal.connect(self.chat_context_extend_for_sending)

    def chat_context_extend(self, sender, **kwargs):
        signal_message = kwargs["message"]
        signal_content = signal_message["content"]
        messages = signal_content["messages"]
        context = signal_content["context"]
        openai_chat = self.app.openai.openai_chat
        
        openai_chat.chat_messages_extend(messages_list=messages, context=context)
        buffer_messages = self.chat_context_to_vectorize_buffer.get(context.id, [])
        buffer_messages.extend(messages)
        self.chat_context_to_vectorize_buffer[context.id] = buffer_messages
        
        _, whether_insert = self.app.openai.auto_bead_insert(context.id)
        
        if whether_insert is False:
            # dashboard display
            model = context.parameters["model"]
            self.dash_board.dash_board_display(tokens_num_window=self.app.get_tokens_window(model))
    
    def chat_context_extend_for_sending(self, sender, **kwargs):
        signal_message = kwargs["message"]
        signal_content = signal_message["content"]
        messages = signal_content["messages"]
        context = signal_content["context"]
        openai_chat = self.app.openai.openai_chat
        
        openai_chat.chat_messages_extend(messages_list=messages, context=context)
        buffer_messages = self.chat_context_to_vectorize_buffer.get(context.id, [])
        buffer_messages.extend(messages)
        self.chat_context_to_vectorize_buffer[context.id] = buffer_messages
        
        self.app.openai.auto_bead_insert(context.id)

    async def chat_context_vectorize(self):
        while self.chat_context_to_vectorize_buffer:
            id, messages_list = self.chat_context_to_vectorize_buffer.popitem()
            self.app.post_message(
                CommonMessage(
                    message_name="vector_memory_write",
                    message_content={
                        "messages_list": messages_list,
                        "context_id": id,
                    }
                )
            )



import logging
import threading

from ..utils.my_text import MyText as Text
from ..utils.my_text import MyLines as Lines
from ..models.signals import response_to_user_message_stream_signal, response_auxiliary_message_signal


gptui_logger = logging.getLogger("gptui_logger")


class ChatResponse:
    def __init__(self, app):
        self.app = app
        self.chat_region = app.main_screen.query_one("#chat_region")
        self.app_chat_tabs = app.main_screen.query_one("#chat_tabs")
        self.buffer = {}
        self.tab_not_switching = threading.Event()
        self.tab_not_switching.set()
        response_to_user_message_stream_signal.connect(self.handle_response)
        response_auxiliary_message_signal.connect(self.handle_group_talk_response)
    
    def delete_buffer_id(self, id: int) -> None:
        self.buffer.pop(id, None)

    def handle_response(self, sender, **kwargs):
        message = kwargs["message"]
        self.stream_display(message)

    def stream_display(self, message: dict, stream: bool = True, copy_code: bool = False) -> None:
        """Display the chat response in TUI"""
        # If the tab is in the process of switching, wait for the chat history to finish loadind
        # before displaying the newly generating chat content.    
        self.tab_not_switching.wait(2)
        context_id = message["content"]["context_id"]
        tab_id = int(self.app_chat_tabs.active[3:])
        if context_id not in self.buffer:
            self.buffer[context_id] = {
                "chat_stream_content": {"role": "assistant", "content": ""},
                "decorate_chat_stream_content_lines": Lines(),
                "last_tab_id": tab_id,
            }
        buffer_context = self.buffer[context_id]
        chat_stream_content = buffer_context["chat_stream_content"]
        char = message["content"]["content"]
        if message["flag"] == "content":
            # This condition being met indicates that the currently generated content corresponds with the active tab window,
            # and it is not the first time being displayed.
            if context_id == tab_id == buffer_context["last_tab_id"]:
                length = len(buffer_context["decorate_chat_stream_content_lines"])
                self.chat_region.right_pop_lines(length, refresh=False)
            chat_stream_content["content"] += char
            if context_id == tab_id:
                buffer_context["decorate_chat_stream_content_lines"] = self.app.decorator(chat_stream_content, stream, copy_code)
                self.chat_region.write_lines(buffer_context["decorate_chat_stream_content_lines"])
        elif message["flag"] == "end":
            if context_id == tab_id:
                self.stream_display(message={"content": {"content": "", "context_id": context_id}, "flag": "content"}, stream=False, copy_code=True)
                self.chat_region.write_lines([Text()])
            chat_stream_content["content"] = ""
            buffer_context["decorate_chat_stream_content_lines"] = Lines()
        buffer_context["last_tab_id"] = tab_id

    def handle_group_talk_response(self, sender, **kwargs):
        message = kwargs["message"]
        message_content = message["content"]
        flag = message["flag"]
        if flag == "group_talk_response":
            self.group_talk_stream_display(message=message_content)
    
    def group_talk_stream_display(self, message: dict, stream: bool = True, copy_code: bool = False) -> None:
        """Display the group talk response in TUI"""
        message_dict = message["content"]
        group_talk_manager_id = message_dict["group_talk_manager_id"]
        tab_id = int(self.app_chat_tabs.active[3:])
        if group_talk_manager_id not in self.buffer:
            self.buffer[group_talk_manager_id] = {
                "group_talk_chat_stream_content": {"role": "assistant", "name": "", "content": ""},
                "group_talk_decorate_chat_stream_content_lines": Lines(),
                "last_tab_id": tab_id,
            }
        buffer_context = self.buffer[group_talk_manager_id]
        group_talk_chat_stream_content = buffer_context["group_talk_chat_stream_content"]
        char = message_dict["content"]
        group_talk_chat_stream_content["name"] = message_dict["name"]
        if message["flag"] == "content":
            # This condition being met indicates that the currently generated content corresponds with the active tab window,
            # and it is not the first time being displayed.
            if group_talk_manager_id == tab_id == buffer_context["last_tab_id"]:
                length = len(buffer_context["group_talk_decorate_chat_stream_content_lines"])
                self.chat_region.right_pop_lines(length, refresh=False)
            group_talk_chat_stream_content["content"] += char
            buffer_context["group_talk_decorate_chat_stream_content_lines"] = self.app.decorator(group_talk_chat_stream_content, stream, copy_code)
            if group_talk_manager_id == tab_id:
                self.chat_region.write_lines(buffer_context["group_talk_decorate_chat_stream_content_lines"])
        elif message["flag"] == "end":
            if group_talk_manager_id == tab_id:
                self.group_talk_stream_display(message={"content": {"role": "assistant", "name": message_dict["name"], "content": "", "group_talk_manager_id": group_talk_manager_id}, "flag": "content"}, stream=False, copy_code=True)
                self.chat_region.write_lines([Text()])
            group_talk_chat_stream_content["content"] = ""
            buffer_context["group_talk_decorate_chat_stream_content_lines"] = Lines()
        buffer_context["last_tab_id"] = tab_id



import argparse
import importlib
import logging
import os
import shutil
import sys
import yaml

from .__init__ import __version__ as gptui_version
from .views.tui import MainApp

APP_VERSION = gptui_version


class ConfigManager:
    
    @staticmethod
    def get_config_path_from_args():
        parser = argparse.ArgumentParser(description="gptui cli")
        parser.add_argument('--config', type=str, help='Path to the configuration file.')
        args = parser.parse_args()
        return os.path.expanduser(args.config) if args.config else None

    @staticmethod
    def copy_default_config_to_user_home():
        default_config_path = importlib.resources.files("gptui") / "config.yml"
        user_config_path = os.path.expanduser('~/.gptui/.config.yml')
        
        if not os.path.exists(user_config_path):
            target_dir = os.path.dirname(user_config_path)
            os.makedirs(target_dir, exist_ok=True)
            shutil.copy(default_config_path, user_config_path)
        return user_config_path

    @staticmethod
    def get_config_path():
        config_path = ConfigManager.get_config_path_from_args()

        if config_path:
            if os.path.exists(config_path):
                return config_path
            else:
                print(f"Config file '{config_path}' dose not exist.")
                sys.exit(1)

        user_config_path = os.path.expanduser('~/.gptui/.config.yml')
        if os.path.exists(user_config_path):
            return user_config_path

        return ConfigManager.copy_default_config_to_user_home()

def gptui():
    config_path = ConfigManager.get_config_path()
    gptui_run(config_path=config_path)
   
def gptui_run(config_path: str) -> None:
    # Retrieve config from config path.
    try:
        with open(os.path.join(os.path.dirname(__file__), '.default_config.yml'), "r") as default_config_file:
            config = yaml.safe_load(default_config_file)
    except FileNotFoundError:
        print(f"Default config file '.default_config.yml' is not found.")
        sys.exit(1)
    try:
        with open(config_path, "r") as config_file:
            user_config = yaml.safe_load(config_file)
    except FileNotFoundError:
        pass
    else:
        config.update(user_config)
    
    log_path = os.path.expanduser(config["log_path"])
    log_level_dict = {
        "DEBUG": logging.DEBUG,
        "INFO": logging.INFO,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
        "CRITICAL": logging.CRITICAL,
    }
    
    log_dir = os.path.dirname(log_path)
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)

    logging.basicConfig(
            filename=log_path,
            filemode='w',
            level=log_level_dict.get(config["log_level"], logging.INFO),
            format="%(asctime)s - %(name)s - %(levelname)s -[%(funcName)s] - %(message)s",
            )
    gptui_logger = logging.getLogger('gptui_logger')

    app = MainApp(config_path, app_version=APP_VERSION)
    reply = app.run()
    if reply:
        print(reply)


if __name__ == "__main__":

    config_path = os.path.join(os.path.dirname(__file__), "config.yml")
    gptui_run(config_path=config_path)



import asyncio
import logging
import time
import os
from collections import deque
from time import monotonic
from typing import cast

import openai
from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.message import Message
from textual.reactive import reactive
from textual.widgets import Button, ProgressBar, Static

from .theme import ThemeColor
from .theme import theme_color as tc
from ..models.utils.openai_api import openai_api
from ..utils.my_text import MyText as Text


gptui_logger = logging.getLogger("gptui_logger")


class TimeDisplay(Static):
    """A widget to display elapsed time."""

    start_time = reactive(monotonic)
    time = reactive(0.0)
    total = reactive(0.0)

    def on_mount(self) -> None:
        self.update_timer = self.set_interval(1 / 9, self.update_time, pause=True)

    def update_time(self) -> None:
        """Method to update time to current."""
        self.time = self.total + (monotonic() - self.start_time)

    def watch_time(self, time: float) -> None:
        """Called when the time attribute changes."""
        minutes, seconds = divmod(time, 60)
        hours, minutes = divmod(minutes, 60)
        self.update(f"{hours:02,.0f}:{minutes:02.0f}:{seconds:05.2f}")

    def start(self) -> None:
        """Method to start (or resume) time updating."""
        self.start_time = monotonic()
        self.update_timer.resume()

    def stop(self):
        """Method to stop the time display updating."""
        self.update_timer.pause()
        self.total += monotonic() - self.start_time
        self.time = self.total

    def reset(self):
        """Method to reset the time display to zero."""
        self.total = 0
        self.time = 0


class Voice(Static, can_focus=True):
    """A stopwatch widget."""

    def __init__(self, app, dot_env_path: str, max_record_time: int = 60):
        super().__init__()
        self.myapp = app
        self.max_record_time = max_record_time
        self.timer = None
        os.makedirs(os.path.join(app.config["workpath"], "temp"), exist_ok=True)
        self.progress_timer = self.set_interval(1, self.progress_drive, pause=True)
        self.key_time_deque = deque(maxlen=2)
        self.voice_record_start_handle = None
        self.openai_api = openai_api(dot_env_path)

    BINDINGS = [
        ("space", "record", "start or end record"),
    ]
    
    DEFAULT_CSS = """
    Voice {
        border: round white;
        layout: horizontal;
        background: $boost;
        height: 1fr;
        padding: 1;
    }
    TimeDisplay {
        content-align: center middle;
        text-opacity: 60%;
        height: 1;
        width: 1fr;
    }
    Button {
        width: 15;
        height: 3;
    }
    #progress_bar {
        width: 100%;
    }
    #stop {
        display: none;
    }
    #voice_status_region{
        width: 100%;
        text-align: center;
    }
    .started {
        text-style: bold;
        background: $success 30%;
        color: $text;
    }
    .started TimeDisplay {
        text-opacity: 100%;
    }
    .started #start {
        display: none
    }
    .started #stop {
        display: block
    }
    .started #send {
        visibility: hidden
    }
    .started #clear {
        visibility: hidden
    }
    """
    
    async def on_button_pressed(self, event: Button.Pressed) -> None:
        """Event handler called when a button is pressed."""
        button_id = event.button.id
        time_display = self.query_one(TimeDisplay)
        
        if button_id == "start":
            self.voice_record_start(self.max_record_time)
            time_display.start()
            self.progress_timer.resume()
            self.query_one("#control_region").add_class("started")
        
        elif button_id == "stop":
            self.voice_record_quit()
            time_display.stop()
            self.progress_timer.pause()
            self.query_one("#control_region").remove_class("started")
            send_button = self.query_one("#send")
            clear_button = self.query_one("#clear")
            send_button.variant = "success"
            clear_button.variant = "success"
            send_button.disabled = False
            clear_button.disabled = False
        
        elif button_id == "send":
            self.progress_timer.pause()
            self.query_one("#voice_status_region").update(Text("Transcribing ...", tc("green") or "green"))
            def blocking_transcribe():
                try:
                    with open(os.path.join(self.myapp.config["workpath"], "temp/voice_temp.wav"), "rb") as audio_file:
                        transcript = self.openai_api.audio.transcriptions.create(
                            model="whisper-1",
                            file=audio_file,
                            response_format="text",
                        )
                except FileNotFoundError:
                    self.query_one("#voice_status_region").update(Text("Have no voice file", tc("red") or "red"))
                    return
                except openai.APIConnectionError as e:
                    self.query_one("#voice_status_region").update(Text(f"APIConnectionError: {e}", tc("red") or "red"))
                    return
                except Exception as e:
                    self.query_one("#voice_status_region").update(Text(f"Unknown error: {e}", tc("red") or "red"))
                    return
                return transcript
            transcript = await asyncio.to_thread(blocking_transcribe)
            self.query_one("#voice_status_region").update(Text("Transcription Completion!", tc("green") or "green"))
            if transcript:
                transcript_text = cast(str, transcript)
                voice_text = transcript_text.strip()
                self.post_message(self.Submitted(voice_text))
                time_display.reset()
                self.query_one("#progress_bar").update(total=self.max_record_time, progress=0)
                send_button = self.query_one("#send")
                clear_button = self.query_one("#clear")
                send_button.variant = "default"
                clear_button.variant = "default"
                send_button.disabled = True
                clear_button.disabled = True
        
        elif button_id == "clear":
            self.progress_timer.pause()
            time_display.reset()
            self.query_one("#progress_bar").update(total=self.max_record_time, progress=0)
            send_button = self.query_one("#send")
            clear_button = self.query_one("#clear")
            send_button.variant = "default"
            clear_button.variant = "default"
            send_button.disabled = True
            clear_button.disabled = True
            self.query_one("#voice_status_region").update(Text("Cleared!", tc("green") or "green"))

    def compose(self) -> ComposeResult:
        """Create child widgets of a stopwatch."""
        with Horizontal(id="control_region"):
            yield Button("Start", id="start", variant="success")
            yield Button("Stop", id="stop", variant = "success" if ThemeColor._theme == "monochrome" else "error")
            with Vertical(id="display_region"):
                yield TimeDisplay()
                yield ProgressBar(total=self.max_record_time, id="progress_bar", show_eta=False)
                yield Static(id="voice_status_region")
            yield Button("Clear", id="clear", disabled=True)
            yield Button("Send", id="send", disabled=True)

    def progress_drive(self):
        self.query_one("#progress_bar").advance(1)

    def voice_record_start(self, time: int = 60) -> None:
        exit_code, self.voice_record_start_handle = self.myapp.drivers.voice_record_start(
            os.path.join(self.myapp.config["workpath"], "temp/voice_temp.wav"),
            time,
        )
        if exit_code:
            self.query_one("#voice_status_region").update(Text(f"Record failed with exit code {exit_code}.", tc("red") or "red"))
        else:
            self.query_one("#voice_status_region").update(Text(f"Recording ...", tc("green") or "green"))

    def voice_record_quit(self) -> None:
        assert self.voice_record_start_handle is not None
        exit_code = self.myapp.drivers.voice_record_quit(self.voice_record_start_handle)
        if exit_code:
            self.query_one("#voice_status_region").update(Text(f"Record quit failed with exit code {exit_code}.", tc("red") or "red"))

    def action_record(self) -> None:
        if self.timer:
            self.timer.reset()
        else:
            self.hot_key_start()
            self.timer = self.set_timer(0.8, self.hot_key_stop)
        self.key_time_deque.append(time.time())

    def hot_key_start(self):
        self.query_one("#start").press()

    async def hot_key_stop(self):
        self.query_one("#stop").press()
        self.timer = None
        await asyncio.sleep(0.6)
        if len(self.key_time_deque) < 2:
            return
        time_interval = self.key_time_deque[1] - self.key_time_deque[0]
        if time_interval > 0.2:
            self.query_one("#clear").press()
        else:
            self.query_one("#send").press()
        self.key_time_deque.clear()

    def on_click(self):
        self.focus()

    class Submitted(Message):
        def __init__(self, content: str) -> None:
            self.content = content
            super().__init__()



from __future__ import annotations
import asyncio
import inspect
import logging
import os
import textwrap
from typing import NamedTuple, Callable, Awaitable, Coroutine, TypeVar, Generic

from rich.console import RenderResult, Console
from rich.style import Style
from rich.text import TextType
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical, Grid
from textual.css.query import NoMatches
from textual.geometry import Size, Offset
from textual.message import Message
from textual.message_pump import _MessagePumpMeta
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import (
    Button,
    DirectoryTree,
    Input,
    Label,
    Markdown,
    RichLog,
    Switch,
    Static,
)

from .screens import SelectPathDialog, MarkdownPreview
from .theme import ThemeColor
from .theme import theme_color as tc
from ..controllers.tube_files_control import TubeFiles
from ..models.doc import Doc
from ..utils.file_icon import file_icon
from ..utils.line_count import my_line_count
from ..utils.my_text import MyText as Text
from ..utils.my_text import MyLines as Lines


gptui_logger = logging.getLogger("gptui_logger")

CheckBoxPointer = TypeVar("CheckBoxPointer")
"""The type of the pointer for a given instance of a [CheckBox]"""

SliderPointer = TypeVar("SliderPointer")
"""The type of the pointer for a given instance of a [Slider]"""


class MyFillIn(Widget):
    """
    Fill area with characters.
    args:
        char: The characters to fill with.
        width: The width of filling area, int or percentage str (int%).
        height: The height of filling area, int or percentage str (int%).
            number refers to the absolute number of rows or columns, and the number% is relative to the container.
        x_start: drawing offsets.
        y_start: drawing offsets.
        color: str of color.
            example: #002CFF, grb(20,200,2)
    """
    def __init__(
        self,
        char: str,
        width: int | str = '100%',
        height: int | str = '100%',
        x_start: int | str = '0%',
        y_start: int | str = '0%',
        color: str | None = None,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.width = width
        self.height = height
        self.char = char
        self.x_start = x_start
        self.y_start = y_start
        self.color = color

    def render(self) -> RenderResult:
        import math
        if isinstance(self.width, int):
            width = self.width
        elif isinstance(self.width, str) and self.width.endswith('%'):
            width = math.floor(self.content_size.width * float(self.width.strip('%')) / 100)
        else:
            raise TypeError(f"Expected a int or a percentage string, bug got{type(self.width).__name__} instead.")
        if isinstance(self.height, int):
            height = self.height
        elif isinstance(self.height, str) and self.height.endswith('%'):
            height = math.floor(self.content_size.height * float(self.height.strip('%')) / 100)
        else:
            raise TypeError(f"Expected a int or a percentage string, bug got{type(self.height).__name__} instead.")
        if isinstance(self.x_start, int):
            x_start = self.x_start
        elif isinstance(self.x_start, str) and self.x_start.endswith('%'):
            x_start = math.floor(self.content_size.width * float(self.x_start.strip('%')) / 100)
        else:
            raise TypeError(f"Expected a int or a percentage string, bug got{type(self.x_start).__name__} instead.")
        if isinstance(self.y_start, int):
            y_start = self.y_start
        elif isinstance(self.y_start, str) and self.y_start.endswith('%'):
            y_start = math.floor(self.content_size.width * float(self.y_start.strip('%')) / 100)
        else:
            raise TypeError(f"Expected a int or a percentage string, bug got{type(self.y_start).__name__} instead.")

        string = '\n' * y_start + (' ' * x_start + self.char * width + '\n') * height
        result = Text(string, style=Style(color=self.color))
        return result


class MyScroll(NamedTuple):
    x_start: float
    x_end: float
    y_start: float
    y_end: float
    def clamped(self):
        x_start, x_end, y_start, y_end = self
        return MyScroll(0 if x_start < 0 else x_end if x_start > x_end else x_start,
                        1 if x_end > 1 else x_start if x_start > x_end else x_end,
                        0 if y_start < 0 else y_end if y_start > y_end else y_start,
                        1 if y_end > 1 else y_start if y_start > y_end else y_end)


class MyScrollSupportMixin(Widget):

    class MyScrollChanged(Message):
        def __init__(self, my_scroll: MyScroll, id: str|None = None) -> None:
            self.my_scroll = my_scroll
            self.id = id
            super().__init__()

    class MyScrollOffsetChanged(Message):
        def __init__(self, my_scroll_offset: Offset, id: str|None = None) -> None:
            self.my_scroll_offset = my_scroll_offset
            self.id = id
            super().__init__()
    
    def init(self):
        self.mouse_move_start_point = False
        self.mouse_move_status = False
    
    my_scroll_offset = reactive(Offset(0,0))
    my_scroll = reactive(MyScroll(0,1,0,1))
    
    def set_scroll(self, x_start: float = 0, x_end: float = 1, y_start: float = 0, y_end: float = 1):
        self.my_scroll = MyScroll(x_start, x_end, y_start, y_end)
    def move_scroll(self, x: float = 0, y: float = 0):
        self.my_scroll = MyScroll(self.my_scroll.x_start + x, self.my_scroll.x_end + x, self.my_scroll.y_start + y, self.my_scroll.y_end + y)

    @property
    def my_scroll_display_size(self) -> Size: 
        return self.my_set_display_size() if callable(self.my_set_display_size) else self.my_set_display_size
    @property
    def my_scroll_virtual_size(self) -> Size:
        return self.my_set_virtual_size() if callable(self.my_set_virtual_size) else self.my_set_virtual_size

    def _my_compute_my_scroll(self) -> MyScroll:
        scroll_x_start = self.my_scroll_offset.x / self.my_scroll_virtual_size.width
        scroll_x_end = (self.my_scroll_offset.x + self.my_scroll_display_size.width) / self.my_scroll_virtual_size.width
        scroll_y_start = self.my_scroll_offset.y / self.my_scroll_virtual_size.height
        scroll_y_end = (self.my_scroll_offset.y + self.my_scroll_display_size.height) / self.my_scroll_virtual_size.height
        return MyScroll(scroll_x_start, scroll_x_end, scroll_y_start, scroll_y_end).clamped()

    def watch_my_scroll_offset(self, new_value) -> None:
        if self.content_size.width != 0:
            self.my_scroll = self._my_compute_my_scroll()
        self.post_message(self.MyScrollOffsetChanged(self.my_scroll_offset, self.id))
    def watch_my_scroll(self, my_scroll) -> None:
        self.post_message(self.MyScrollChanged(self.my_scroll, self.id))

    def my_scroll_content_to_render(self, content_wrap: Lines, y: int = 0, h: int = 0, x: int = 0, w: int = 0) -> Text:
        result = Text()
        line_nums = len(content_wrap)
        start_line = min(y, line_nums)
        end_line = min(y+h, line_nums)
        for line in content_wrap[start_line:end_line]:
            result.append_text(line)
            result.append_text(Text('\n'))
        return result

# 鼠标拖动的滚动支持
#    def on_mouse_move(self, event: events.MouseMove) -> None:
#        if self.mouse_move_start_point and 0 <= event.x <= self.my_scroll_display_size.width and 0 <= event.y <= self.my_scroll_display_size.height:
#            if not self.mouse_move_status:
#                delta_x = 0
#                delta_y = 0
#            else:
#                delta_x = event.delta_x
#                delta_y = event.delta_y
#            self.mouse_move_status = True
#            if self.my_scroll_virtual_size.width > self.my_scroll_display_size.width:
#                self.my_scroll_offset = Offset(self.my_scroll_offset.x - delta_x, self.my_scroll_offset.y)
#                if self.my_scroll_offset.x < 0:
#                    self.my_scroll_offset = Offset(0, self.my_scroll_offset.y)
#                if self.my_scroll_offset.x > self.my_scroll_virtual_size.width - self.my_scroll_display_size.width:
#                    self.my_scroll_offset = Offset(self.my_scroll_virtual_size.width - self.my_scroll_display_size.width, self.my_scroll_offset.y)
#            else:
#                self.my_scroll_offset = Offset(0, self.my_scroll_offset.y)
#            if self.my_scroll_virtual_size.height > self.my_scroll_display_size.height:
#                self.my_scroll_offset = Offset(self.my_scroll_offset.x, self.my_scroll_offset.y - delta_y)
#                if self.my_scroll_offset.y < 0:
#                    self.my_scroll_offset = Offset(self.my_scroll_offset.x, 0)
#                if self.my_scroll_offset.y > self.my_scroll_virtual_size.height - self.my_scroll_display_size.height:
#                    self.my_scroll_offset = Offset(self.my_scroll_offset.x, self.my_scroll_virtual_size.height - self.my_scroll_display_size.height)
#            else:
#                self.my_scroll_offset = Offset(self.my_scroll_offset.x, 0)
#    
#    def on_mouse_up(self, event) -> None:
#        self.mouse_move_status = False
#        self.mouse_move_start_point = False
#    def on_mouse_down(self, event) -> None:
#        if 0 <= event.x <= self.my_scroll_display_size.width and 0 <= event.y <= self.my_scroll_display_size.height:
#            self.mouse_move_start_point = True

    def on_mouse_scroll_up(self, event) -> None:
        if self.my_scroll_virtual_size.height > self.my_scroll_display_size.height:
            self.my_scroll_offset = Offset(self.my_scroll_offset.x, self.my_scroll_offset.y - 1)
            if self.my_scroll_offset.y < 0:
                self.my_scroll_offset = Offset(self.my_scroll_offset.x, 0)
            if self.my_scroll_offset.y > self.my_scroll_virtual_size.height - self.my_scroll_display_size.height:
                self.my_scroll_offset = Offset(self.my_scroll_offset.x, self.my_scroll_virtual_size.height - self.my_scroll_display_size.height)
        else:
            self.my_scroll_offset = Offset(self.my_scroll_offset.x, 0)
    def on_mouse_scroll_down(self, event) -> None:
        if self.my_scroll_virtual_size.height > self.my_scroll_display_size.height:
            self.my_scroll_offset = Offset(self.my_scroll_offset.x, self.my_scroll_offset.y + 1)
            if self.my_scroll_offset.y < 0:
                self.my_scroll_offset = Offset(self.my_scroll_offset.x, 0)
            if self.my_scroll_offset.y > self.my_scroll_virtual_size.height - self.my_scroll_display_size.height:
                self.my_scroll_offset = Offset(self.my_scroll_offset.x, self.my_scroll_virtual_size.height - self.my_scroll_display_size.height)
        else:
            self.my_scroll_offset = Offset(self.my_scroll_offset.x, 0)


class MyMultiInput(Input, MyScrollSupportMixin):
    """
    use "on_input_submitted" method to handle the submitting event.
    """
    # TODO: uncompleted
    Input.BINDINGS = [
        Binding("left", "cursor_left", "cursor left", show=False),
        Binding("ctrl+left", "cursor_left_word", "cursor left word", show=False),
        Binding("right", "cursor_right", "cursor right", show=False),
        Binding("ctrl+right", "cursor_right_word", "cursor right word", show=False),
        Binding("backspace", "delete_left", "delete left", show=False),
        Binding("home,ctrl+a", "home", "home", show=False),
        Binding("end,ctrl+e", "end", "end", show=False),
        Binding("delete,ctrl+d", "delete_right", "delete right", show=False),
        Binding("enter", "submit", "submit", show=False),
        Binding(
            "ctrl+w", "delete_left_word", "delete left to start of word", show=False
        ),
        Binding("ctrl+u", "delete_left_all", "delete all to the left", show=False),
        Binding(
            "ctrl+f", "delete_right_word", "delete right to start of word", show=False
        ),
        Binding("ctrl+k", "delete_right_all", "delete all to the right", show=False),
        Binding("up", "cursor_up", "cursor up"),
        Binding("down", "cursor_down", "cursor down"),
        Binding("shift+left", "change_line", "change line"),
    ]   

#    def __init__(self,*args,**kwargs):
#        Input.__init__(self,*args,**kwargs)
#        MyScrollSupportMixin.__init__(self)
    def on_mount(self):
        MyScrollSupportMixin.init(self)
        #self.capture_mouse()
        self.my_set_display_size = self.display_size_send
        self.my_set_virtual_size = self.virtual_size_send

    def virtual_size_send(self):
        width = self.content_size.width
        height = my_line_count(self._value, width)
        return Size(width, height)
    def display_size_send(self):
        return self.content_size

    def render(self) -> RenderResult:
        if not self.value:
            placeholder = Text(self.placeholder, justify="left")
            placeholder.stylize(self.get_component_rich_style("input--placeholder"))
            if self.has_focus:
                cursor_style = self.get_component_rich_style("input--cursor")
                if self._cursor_visible:
                    # If the placeholder is empty, there's no characters to stylise
                    # to make the cursor flash, so use a single space character
                    if len(placeholder) == 0:
                        placeholder = Text(" ")
                    placeholder.stylize(cursor_style, 0, 1)
            return placeholder
        return self._my_render_content(self, self._cursor_visible)

    def _my_render_content(self, input, cursor_visible) -> RenderResult:
        content = input._value
        scroll_x, scroll_y = input.my_scroll_offset
        if input._cursor_at_end:
            content.pad_right(1)
        cursor_style = input.get_component_rich_style("input--cursor")
        if cursor_visible and input.has_focus:
            cursor = input.cursor_position
            content.stylize(cursor_style, cursor, cursor+1)
        width = input.content_size.width
        height = input.content_size.height
        length = content.cell_len
        content_wrap = content.wrap(console=Console(), width=self.content_size.width, overflow="fold", no_wrap=False)
        return self.my_scroll_content_to_render(content_wrap, y=scroll_y, h=height)

    def action_cursor_up(self) -> None:
        self.cursor_position -= self.content_size.width
    def action_cursor_down(self) -> None:
        self.cursor_position += self.content_size.width
    def action_change_line(self) -> None:
        self.value = self.value + "\n"
        self.cursor_position += self.content_size.width - self.cursor_position % self.content_size.width
    def action_submit(self):
        self.post_message(self.Submitted(self, self.value))
        self.value = ""


class MyChatWindow(MyScrollSupportMixin):
    """
    supplied two write content method:
    1. update, write, right_crop
    2. update_lines, write_lines, right_pop_lines
    methods with line is recomended
    Important: always use only one serial methods, mix use will cause error
    """
    def on_mount(self):
        MyScrollSupportMixin.init(self)
        #self.capture_mouse()
        self.my_set_display_size = self.display_size_send
        self.my_set_virtual_size = self.virtual_size_send
        self.my_content = Text()
        self.my_content_wrap = Lines()
        self.refresh_content_wrap_request = False
        self.right_crop_request = 0
        self.right_crop_request = 0

    def virtual_size_send(self):
        width = self.content_size.width
        height = len(self.my_content_wrap)
        return Size(width, height)
    
    def display_size_send(self):
        return self.content_size

    def clear(self, refresh: bool = True):
        self.my_content = Text()
        self.my_content_wrap = Lines()
        self.right_crop_request = 0
        self.right_pop_lines_request = 0
        self.refresh_content_wrap_request = False
        if refresh:
            self.refresh()

    def update(self, content: Text, scroll_to_end: bool = True) -> None:
        self.my_content = content
        self.my_content_wrap = self.my_content.wrap(console=Console(), width=self.content_size.width, overflow="fold", no_wrap=False)
        if scroll_to_end:
            self.scroll_to_end()
        self.refresh()
        self.right_crop_request = 0
        self.refresh_content_wrap_request = False
    
    def write(self, content: Text, scroll_to_end: bool = True) -> None:
        console = Console()
        width = self.content_size.width
        if self.right_crop_request:
            self.my_content.right_crop(self.right_crop_request) # right_crop(0) will crop all characters, so should be skipped
            self.right_crop_request = 0
            self.my_content.append_text(content)
            self.my_content_wrap = self.my_content.wrap(console=console, width=width, overflow="fold", no_wrap=False)
            self.refresh_content_wrap_request = False
        else:
            if self.refresh_content_wrap_request is True:
                self.refresh_content_wrap()
                self.refresh_content_wrap_request = False
            self.my_content.append_text(content)
            if self.my_content_wrap:
                last_line = self.my_content_wrap.pop()
            else:
                last_line = Text()
            last_line.append_text(content)
            last_line_after = last_line.wrap(console=console, width=width, overflow="fold", no_wrap=False)
            self.my_content_wrap.extend(last_line_after)
        if scroll_to_end:
            self.scroll_to_end()
        self.refresh()

    def scroll_to_end(self, refresh: bool = False):
        if self.refresh_content_wrap_request is True:
            self.refresh_content_wrap()
            self.refresh_content_wrap_request = False
        offset_y = self.my_scroll_virtual_size.height - self.my_scroll_display_size.height
        self.my_scroll_offset = Offset(self.my_scroll_offset.x, offset_y if offset_y > 0 else 0)
        if refresh is True:
            self.refresh()
    
    def right_crop(self, amount: int, refresh: bool = True, scroll_to_end: bool = False) -> None:
        if refresh:
            amount += self.right_crop_request
            if amount:
                self.my_content.right_crop(amount) # right_crop(0) will crop all characters, so should be skipped
                self.my_content_wrap = self.my_content.wrap(console=Console(), width=self.content_size.width, overflow="fold", no_wrap=False)
            if scroll_to_end:
                self.scroll_to_end()
            self.refresh()
            self.right_crop_request = 0
            self.refresh_content_wrap_request = False
        else:
            self.right_crop_request += amount

    def write_content_without_display(self, content: Text):
        self.my_content.append_text(content)
        self.refresh_content_wrap_request = True

    def refresh_content_wrap_request_execute(self):
        if self.refresh_content_wrap_request is True:
            self.refresh_content_wrap()
            self.refresh_content_wrap_request = False

    def refresh_content_wrap(self):
        if self.my_content:
            self.my_content_wrap = self.my_content.wrap(console=Console(), width=self.content_size.width, overflow="fold", no_wrap=False)
        else:
            self.my_content_wrap = Lines()

    def update_lines(self, content_lines: Lines, scroll_to_end: bool = True) -> None:
        self.my_content_wrap = content_lines
        if scroll_to_end:
            self.scroll_to_end()
        self.refresh()
        self.right_pop_lines_request = 0

    def write_lines(self, content_lines: Lines, scroll_to_end: bool = True) -> None:
        if self.right_pop_lines_request:
            num = min(self.right_pop_lines_request, len(self.my_content_wrap))
            for _ in range(num):
                self.my_content_wrap.pop()
            self.right_pop_lines_request = 0
        self.my_content_wrap.extend(content_lines)
        if scroll_to_end:
            self.scroll_to_end()
        self.refresh()

    def right_pop_lines(self, amount: int, refresh: bool = True, scroll_to_end: bool = False) -> list[Text]:
        if refresh:
            amount += self.right_pop_lines_request
            out = []
            amount = min(amount, len(self.my_content_wrap))
            for _ in range(amount):
                out.append(self.my_content_wrap.pop())
            if scroll_to_end:
                self.scroll_to_end()
            self.refresh()
            return out.reverse()
        else:
            self.right_pop_lines_request += amount
            return []

    @property
    def my_render_content(self) -> Text:
        scroll_x, scroll_y = self.my_scroll_offset
        return self.my_scroll_content_to_render(self.my_content_wrap, y = scroll_y, h = self.my_scroll_display_size.height)

    def render(self) -> RenderResult:
        result = self.my_render_content
        return result

    class Resize(Message):
        def __init__(self, event):
            self.event = event
            super().__init__()

    def on_resize(self, event) -> None:
        self.post_message(self.Resize(event))


class _InputDialog(Widget):
    """This is a awaitable input dialog. May be useful."""
    # Not used yet, switched to using InputDialog by Screen. Kept as future reference.
    """
    It should be run in a worker!
    """
    
    DEFAULT_CSS = """
    _InputDialog{
        width: 50;
        height: 12;
        border: round;
    }
    Input{
        height: 1;
    }
    Label{
        border: none;
    }
    Button{
        width: 12;
        min-width: 6;
        min-height: 1;
    }
    """
    def __init__(self, prompt: str, *args, **kwargs) -> None:
        self.prompt = prompt
        self.status = False
        self.value: str|None = ''
        super().__init__(*args, **kwargs)

    def on_mount(self) -> None:
        self.border_title = "Input Dialog"
        self.query_one("#label").styles.border = None
        
    def compose(self) -> ComposeResult:
        yield Label(self.prompt, id = "label")
        yield Input()
        with Horizontal():
            yield Button("OK")
            yield Button("Default")
            yield Button("Cancel")

    def on_button_pressed(self, event) -> None:
        if event.button.label == "OK":
            self.value = self.query_one(Input).value
        elif event.button.label == "Cancel":
            self.value = None
        else:
            self.value = self.query_one(Input).value
        self.status = True

    @classmethod
    async def my_input_dialog(cls, app, prompt: str) -> str|None:
        dialog_instance = _MyInputDialog(prompt, classes = "top_layer")
        await app.mount(dialog_instance)
        while not dialog_instance.status:
            await asyncio.sleep(0.1)
        result = dialog_instance.value
        dialog_instance.remove()
        return result


class Tube(Widget):
    DEFAULT_CSS = """
    Tube Horizontal{
        height: 1;
    }
    Tube Button{
        border: none;
        width: 10;
        min-width: 5;
        height: 1;
        min-height: 1;
        padding: 0;
        margin: 0;
    }
    #up_control {
        height: 1;
        grid-size: 3 1;
        grid-columns: 10 1fr 4;
    }
    Tube GridContentUp {
        background: rgb(30, 35, 50);
    }
    Tube GridContentDown {
        background: rgb(30, 35, 50);
    }
    #up_tube{
        height: 1fr;
    }
    #down_tube{
        height: 1fr;
    }
    """

    def __init__(self, app, **kwargs):
        self.myapp = app
        super().__init__(**kwargs)
        self.up_tube_list = []
        self.down_tube_list = []
        self.export_content = None

    def compose(self) -> ComposeResult:
        with Vertical():
            with Grid(id="up_control"):
                yield Button("|Import|", id="up_import")
                yield Button("|Clear|", id="up_clear")
                yield Switch(value=False, id="send_switch", classes="min_switch")
            yield GridContentMake(name="GridContentUp", id="up_tube", column_num=4, grid_rows="5")
            with Horizontal():
                yield Button("|Export|", id="down_export")
                yield Button("|Delete|", id="down_delete")
                yield Button("|Clear|", id ="down_clear")
            yield GridContentMake(name="GridContentDown", id="down_tube", column_num=4, grid_rows="5")

    def on_mount(self):
        export_button = self.query_one("#down_export")
        delete_button = self.query_one("#down_delete")
        export_button.can_focus = False
        delete_button.can_focus = False

    def add_document_to_up_tube(self, document: Doc) -> None:
        self.up_tube_list.append(document)
        document_name = document.name
        file_type = document.ext
        file_description = document_name + file_type
        up_tube = self.query_one("#up_tube")
        up_tube.add_children(
            FileIcon(
                pointer=document,
                file_type=file_type,
                file_label=document_name,
                file_description=file_description,
                previewer=self.myapp,
            )
        )

    def add_document_to_down_tube(self, document: Doc) -> None:
        self.down_tube_list.append(document)
        doc_ext = document.ext
        doc_name = document.name
        down_tube = self.query_one("#down_tube")
        down_tube.add_children(
            FileIcon(
                pointer=document,
                file_type=doc_ext,
                file_label=doc_name,
                file_description = doc_name + doc_ext,
                previewer=self.myapp,
            )
        )
    
    async def on_button_pressed(self, event: Button.Pressed) -> None:
        button_id = event.button.id
        if button_id == "up_import":
            self.myapp.main_screen.query_one("#middle_switch").change_to_pointer("directory_tree")
        if button_id == "up_clear":
            self.query_one("#up_tube").clear()
            self.up_tube_list = []
        if button_id == "down_export":
            try:
                focused_file = self.query("#down_tube > FileIcon:focus").first()
            except NoMatches:
                self.myapp.main_screen.query_one("#status_region").update(Text("No file is selected.", tc("yellow") or "yellow"))
            except Exception as e:
                self.myapp.main_screen.query_one("#status_region").update(Text(f"{e}", tc("yellow") or "yellow"))
            else:
                self.export_content = focused_file.pointer.content
                self.myapp.push_screen(
                    SelectPathDialog(
                        root_directory_path=self.myapp.config["directory_tree_path"],
                        prompt="Determine the file path and name (including extension):",
                        placeholder="filename"
                    ),
                    self._save_file,
                )
        if button_id == "down_clear":
            self.query_one("#down_tube").clear()
            self.down_tube_list = []
        if button_id == "down_delete":
            try:
                focused_file = self.query("#down_tube > FileIcon:focus").first()
            except NoMatches:
                self.myapp.main_screen.query_one("#status_region").update(Text("No file is selected.", tc("yellow") or "yellow"))
            except Exception as e:
                self.myapp.main_screen.query_one("#status_region").update(Text(f"{e}", tc("yellow") or "yellow"))
            else:
                self.query_one("#down_tube").remove_child(focused_file)
                self.down_tube_list.remove(focused_file.pointer)

    async def _save_file(self, selected_path: tuple[bool, str]) -> None:
        status, path = selected_path
        if status is False:
            return
        status_region = self.myapp.main_screen.query_one("#status_region")
        
        # Check whether it includes a path name.
        dirname, filename = os.path.split(path)
        if not dirname:
            path = os.path.join(self.myapp.config["workpath"], filename)

        if os.path.isfile(path):
            status_region.update(Text("File is not exported beacause file name have existed!", tc("red") or "red"))
            return
        if not self.export_content:
            status_region.update(Text("File is empty! Not exported.", tc("yellow") or "yellow"))
            return
        tf = TubeFiles(status_region)
        await tf.write_file_async(file_path=path, file_content=self.export_content)

    def get_upload_files(self) -> list[Doc]:
        return self.up_tube_list

    def refresh_display(self) -> None:
        up_tube_list = self.up_tube_list.copy()
        down_tube_list = self.down_tube_list.copy()
        self.query_one("#up_tube").clear()
        self.query_one("#down_tube").clear()
        self.up_tube_list = []
        self.down_tube_list = []
        for doc in up_tube_list:
            self.add_document_to_up_tube(doc)
        for doc in down_tube_list:
            self.add_document_to_down_tube(doc)


class MultiGridContent(Widget):
    DEFAULT_CSS = """
    MultiGridContent > * {
        height: 8;
        border: none;
        background: rgb(30, 35, 50);
    }
    MultiGridContent MyFillIn {
        height: 1;
        color: rgb(100,100,100);
        background: rgb(30, 35, 50);
        opacity: 0.5;
    }
    MultiGridContent {
        background: rgb(30, 35, 50);
    }
    """
    
    def __init__(self, grid_list: list, **kwargs):
        self.init_grid_list = grid_list
        super().__init__(**kwargs)

    def compose(self) -> ComposeResult:
        yield self.init_grid_list[0]
        for grid_content in self.init_grid_list[1:]:
            yield MyFillIn(char="\u2500", width="80%", x_start="10%")
            yield grid_content

    def add_grid_content(
        self,
        *grid_contents,
        before: int | str | Widget | None = None,
        after: int | str | Widget | None = None
    ) -> None:
        self.mount(*grid_contents, before=before, after=after)

    def add_horizontal_line(
        self,
        char: str,
        before: int | str | Widget | None = None,
        after: int | str | Widget | None = None,
    ) -> None:
        self.mount(MyFillIn(char), before=before, after=after)
    
    @property
    def grid_content_list(self) -> list:
        children = self.query("MultiGridContent > GridContentBase")
        return list(children)

    @property
    def horizontal_line_list(self) -> list:
        children = self.query("MultiGridContent > MyFillIn")
        return list(children)
    
    @property
    def children_list(self) -> list:
        children = self.query("MultiGridContent > *")
        return list(children)

    def remove_grid_content(self, child: int | Widget) -> None:
        if isinstance(child, int):
            try:
                grid_content = self.grid_content_list[child]
                grid_content.remove()
                position = self.children.index(grid_content)
                if position < len(self.children_list) - 1:
                    if isinstance(self.children_list[position + 1], MyFillIn):
                        self.children_list[position + 1].remove()
            except IndexError as e:
                gptui_logger.error(e)
        else:
            child.remove()
            position = self.children_list.index(child)
            if position < len(self.children_list) - 1:
                if isinstance(self.children_list[position + 1], MyFillIn):
                    self.children_list[position + 1].remove()
    
    def remove_horizontal_line(self, line: int | MyFillIn) -> None:
        if isinstance(line, int):
            try:
                my_fill_in = self.horizontal_line_list[line]
                my_fill_in.remove()
            except IndexError as e:
                gptui_logger.error(e)
        else:
            line.remove()

    def clear(self):
        self.remove_children()


class GridContentMeta(_MessagePumpMeta):
    def __new__(mcs, name, bases, attrs):

        def __init__(
            self, 
            column_num: int, 
            row_num: int | None = None, 
            grid_columns: str = "1fr", 
            grid_rows: str = "2", 
            **kwargs
        ):
            self.column_num = column_num
            self.row_num = row_num
            super(self.__class__, self).__init__(**kwargs)
            # set the grid layout
            self.__class__.DEFAULT_CSS += f"{name} {{layout: grid; grid-size:{column_num} {row_num if row_num else ''}; grid-columns: {grid_columns}; grid-rows: {grid_rows}; grid-gutter: 1;}}"
        
        def add_children(
            self,
            *children,
            before: int | str | Widget | None = None,
            after: int | str | Widget | None = None
        ) -> None:
            self.mount(*children, before=before, after=after)

        def update_children(self, *args, **kwargs) -> None:
            self.clear()
            self.add_children(*args, **kwargs)
        
        def remove_child(self, child: str | Widget) -> None:
            if isinstance(child, str):
                try:
                    self.query_one(child).remove()
                except NoMatches as e:
                    gptui_logger.error(e)
            else:
                child.remove()
        
        def remove_child_last(self) -> None:
            try:
                children = self.query("GridContentBase > *")
                children.last().remove()
            except NoMatches:
                return
        
        def remove_child_first(self):
            try:
                children = self.query("GridContentBase > *")
                children.first().remove()
            except NoMatches:
                return
        
        def clear(self):
            self.remove_children()

        def change_child_order(
            self,
            child,
            before: int | Widget | None = None,
            after: int | Widget | None = None
        ) -> None:
            self.move_child(child, before=before, after=after)
        
        attrs["DEFAULT_CSS"] = f"""
        .box {{
            height: 100%;
            border: solid {tc("green") or "green"};
        }}
        """
        attrs["__init__"] = __init__
        attrs["add_children"] = add_children
        attrs["update_children"] = update_children
        attrs["remove_child"] = remove_child
        attrs["remove_child_last"] = remove_child_last
        attrs["remove_child_first"] = remove_child_first
        attrs["clear"] = clear
        attrs["change_child_order"] = change_child_order

        return type.__new__(mcs, name, bases, attrs)


class GridContentBase(Widget):
    """Base class for GridContent and GridContentMake

    Each GridContentMake must be assigned a unique class name upon instantiation;
    therefore, the name of each GridContentMake is actually distinct and uncertain.
    This base class can be used to determine if a class is a GridContentMake or GridContent.
    """

class GridContentMake:
    """
    A container that put its contents in a grid.
    """
    def __new__(
        cls,
        name: str,
        column_num: int, 
        row_num: int | None = None, 
        grid_columns: str = "1fr", 
        grid_rows: str = "2", 
        **kwargs
    ):
        return GridContentMeta(name, (GridContentBase,), {})(column_num=column_num, row_num=row_num, grid_columns=grid_columns, grid_rows=grid_rows, **kwargs)


class GridContent(GridContentBase):

    DEFAULT_CSS = """
    GridContent {
        layout: grid;
        grid-size: 4;
        grid-columns: 1fr;
        grid-rows: 2;
        grid-gutter: 1;
    }
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def add_children(
        self,
        *children,
        before: int | str | Widget | None = None,
        after: int | str | Widget | None = None
    ) -> None:
        self.mount(*children, before=before, after=after)

    def update_children(self, *args, **kwargs) -> None:
        self.clear()
        self.add_children(*args, **kwargs)
    
    def remove_child(self, child: str | Widget) -> None:
        if isinstance(child, str):
            try:
                self.query_one(child).remove()
            except NoMatches as e:
                gptui_logger.error(e)
        else:
            child.remove()
    
    def remove_child_last(self) -> None:
        try:
            children = self.query("GridContentBase > *")
            children.last().remove()
        except NoMatches:
            return
    
    def remove_child_first(self):
        try:
            children = self.query("GridContentBase > *")
            children.first().remove()
        except NoMatches:
            return
    
    def clear(self):
        self.remove_children()

    def change_child_order(
        self,
        child,
        before: int | Widget | None = None,
        after: int | Widget | None = None
    ) -> None:
        self.move_child(child, before=before, after=after)


class MyCheckBox(Widget, Generic[CheckBoxPointer]):
    def __init__(self, status: bool, icon: Text, label: Text, pointer: CheckBoxPointer, domain=None, **kwargs):
        super().__init__(**kwargs)
        self.status = status
        self.icon = icon
        self.label = label
        self.pointer = pointer
        self.domain = domain

    def compose(self) -> ComposeResult:
        yield self.IconRegion(self.status, self.icon, self)
        yield self.LabelRegion(self.label, self)

    class IconRegion(Static):
        def __init__(self, status: bool, icon: Text, check_box_instance: MyCheckBox):
            super().__init__()
            self.status = status
            self.icon = icon
            self.check_box = check_box_instance
    
        def on_mount(self) -> None:
            self.display()

        def display(self):
            icon_display = Text()
            if self.status is True:
                if ThemeColor._theme == "monochrome":
                    icon_display.append_text(Text("◉", tc("green") or "green"))
                else:
                    icon_display.append_text(Text(u"\u2705"))
            else:
                if ThemeColor._theme == "monochrome":
                    icon_display.append_text(Text("○", tc("green") or "green"))
                else:
                    icon_display.append_text(Text(u"\u2B1C"))
            icon_display.append_text(self.icon)
            self.update(icon_display)
        
        def on_click(self):
            self.status = not self.status
            self.post_message(self.CheckBoxStatusChanged(self.status, self.check_box))
            self.display()
        
        class CheckBoxStatusChanged(Message):
            def __init__(self, status: bool, check_box: MyCheckBox) -> None:
                self.check_box = check_box
                self.status = status
                super().__init__()

    class LabelRegion(Static):
        def __init__(self, label: Text, check_box_instance: MyCheckBox):
            super().__init__()
            self.update(label)
            self.check_box = check_box_instance

        def on_click(self):
            self.post_message(self.CheckBoxLabelClicked(self.check_box))
        
        class CheckBoxLabelClicked(Message):
            def __init__(self, check_box: MyCheckBox) -> None:
                self.check_box = check_box
                super().__init__()


class FileIcon(Static, can_focus=True):
    
    DEFAULT_CSS = """
    FileIcon {
        width: 6;
    }
    FileIcon:focus {
        background: blue;
    }
    """

    BINDINGS = [
        ("space", "preview", "Preview the content of the file"),
    ]
    
    def __init__(self, pointer: Doc, file_label: str, file_type: str, file_description: str, previewer, **kwargs):
        super().__init__(**kwargs)
        icon_display = file_icon(file_label=file_label, file_type=file_type, file_description=file_description)
        self.update(icon_display)
        self.mydisplay = icon_display
        self.pointer = pointer
        self.previewer = previewer
    
    async def on_click(self):
        self.post_message(self.FileIconClicked(self, self.pointer))
        self.styles.opacity = 0.5
        self.styles.animate("opacity", value=1.0, duration=0.5)

    def action_preview(self):
        doc = self.pointer
        content = doc.content
        self.app.push_screen(MarkdownPreview(content, previewer_title=doc.name + doc.ext))

    class FileIconClicked(Message):
        def __init__(self, icon, pointer: Doc) -> None:
            self.icon = icon
            self.pointer = pointer
            super().__init__()


class SlideSwitch(Widget):
    DEFAULT_CSS = """
    SliderSwitcher {
        width: 3;
    }
    """

    def __init__(self, sliders: list, index: int = 0, direction: str = "down", **kwargs):
        super().__init__(**kwargs)
        self.direction = direction
        self.index = index
        self.sliders = []
        for slider in sliders:
            self.sliders.append(self.Slider(slider[0], slider[1], self))
        self.set_sliders(index, refresh=False)

    def compose(self) -> ComposeResult:
        for slider in self.sliders:
            yield slider

    def on_slider_slider_changed(self, event):
        index = event.index
        self.index = index
        self.set_sliders(index)
    
    def change_to_index(self, index):
        slider = self.sliders[index]
        slider.on_click()

    def change_to_pointer(self, aim_pointer) -> None:
        index_list = self.query_pointer(aim_pointer)
        if len(index_list) != 1:
            gptui_logger.error("Match zero or more than one pointer.")
            return
        self.change_to_index(index_list[0])

    def query_pointer(self, aim_pointer) -> list:
        pointers_list = [slider.pointer for slider in self.sliders]
        result = [index for index, pointer in enumerate(pointers_list) if pointer == aim_pointer]
        return result

    def set_sliders(self, index, refresh: bool = True):
        if self.direction == "down":
            if index == 0:
                self.sliders[0].set_status("on_down_first", refresh=refresh)
                for slider in self.sliders[1:-1]:
                    slider.set_status("lower_down", refresh=refresh)
                self.sliders[-1].set_status("off_down_last", refresh=refresh)
            elif index == len(self.sliders) - 1:
                self.sliders[0].set_status("off_down_first", refresh=refresh)
                for slider in self.sliders[1:-1]:
                    slider.set_status("upper_down", refresh=refresh)
                self.sliders[-1].set_status("on_down_last", refresh=refresh)
            else:
                self.sliders[0].set_status("off_down_first", refresh=refresh)
                for slider in self.sliders[1:index]:
                    slider.set_status("upper_down", refresh=refresh)
                self.sliders[index].set_status("on_down", refresh=refresh)
                for slider in self.sliders[index+1:-1]:
                    slider.set_status("lower_down", refresh=refresh)
                self.sliders[-1].set_status("off_down_last", refresh=refresh)
        else:
            if index == 0:
                self.sliders[0].set_status("on_up_first", refresh=refresh)
                for slider in self.sliders[1:-1]:
                    slider.set_status("lower_up", refresh=refresh)
                self.sliders[-1].set_status("off_up_last", refresh=refresh)
            elif index == len(self.sliders) - 1:
                self.sliders[0].set_status("off_up_first", refresh=refresh)
                for slider in self.sliders[1:-1]:
                    slider.set_status("upper_up", refresh=refresh)
                self.sliders[-1].set_status("on_up_last", refresh=refresh)
            else:
                self.sliders[0].set_status("off_up_first", refresh=refresh)
                for slider in self.sliders[1:index]:
                    slider.set_status("upper_up", refresh=refresh)
                self.sliders[index].set_status("on_up", refresh=refresh)
                for slider in self.sliders[index+1:-1]:
                    slider.set_status("lower_up", refresh=refresh)
                self.sliders[-1].set_status("off_up_last", refresh=refresh)
    

    class Slider(Static, Generic[SliderPointer]):
        DEFAULT_CSS = """
        Slider {
            width: 3;
            height: 1;
        }
        """

        def __init__(self, label: Text | str, pointer: SliderPointer, slide_switch):
            super().__init__()
            self.status = None
            self.label = label
            self.pointer = pointer
            self.slide_switch = slide_switch

        def on_mount(self):
            self.display()

        def on_click(self):
            index = self.slide_switch.sliders.index(self)
            self.post_message(self.SliderChanged(self, self.pointer, index, self.slide_switch))

        def display(self):
            if isinstance(self.label, str):
                display_content = Text(self.label).copy()
            else:
                display_content = self.label.copy()
            if self.status in ("on_down", "on_up", "on_down_last", "on_down_first", "on_up_last", "on_up_first"):
                display_content.stylize(tc("red") or "red")
            if self.status == "on_down":
                display_content.append_text(Text("\u2501\u2513", tc("red") or "red"))
            elif self.status == "upper_down":
                display_content.append_text(Text("\u2574\u2502", tc("green") or "green"))
            elif self.status == "lower_down":
                display_content.append_text(Text("\u2574", tc("green") or "green"))
                display_content.append_text(Text("\u2503", tc("red") or "red"))
            elif self.status == "on_up":
                display_content.append_text(Text("\u2501\u251b", tc("red") or "red"))
            elif self.status == "upper_up":
                display_content.append_text(Text("\u2574", tc("green") or "green"))
                display_content.append_text(Text("\u2503", tc("red") or "red"))
            elif self.status == "lower_up":
                display_content.append_text(Text("\u2574\u2502", tc("green") or "green"))
            elif self.status == "on_down_last":
                display_content.append_text(Text("\u2501\u2501", tc("red") or "red"))
            elif self.status == "off_down_last":
                display_content.append_text(Text("\u2574", tc("green") or "green"))
                display_content.append_text(Text("\u2517", tc("red") or "red"))
            elif self.status == "on_down_first":
                display_content.append_text(Text("\u2501\u2513", tc("red") or "red"))
            elif self.status == "off_down_first":
                display_content.append_text(Text("\u2574\u2577", tc("green") or "green"))
            elif self.status == "on_up_first":
                display_content.append_text(Text("\u2501\u2501", tc("red") or "red"))
            elif self.status == "off_up_first":
                display_content.append_text(Text("\u2574", tc("green") or "green"))
                display_content.append_text(Text("\u250f", tc("red") or "red"))
            elif self.status == "on_up_last":
                display_content.append_text(Text("\u2501\u251b", tc("red") or "red"))
            elif self.status == "off_up_last":
                display_content.append_text(Text("\u2574\u2575", tc("green") or "green"))
            self.update(display_content)

        def set_status(self, status, refresh: bool = True):
            self.status = status
            if refresh is True:
                self.display()

        class SliderChanged(Message):
            def __init__(self, slider, pointer, index, slide_switch) -> None:
                self.slider = slider
                self.pointer = pointer
                self.index = index
                self.slide_switch = slide_switch
                super().__init__()


class MyMarkdownWindow(Widget, can_focus=True):
    
    DEFAULT_CSS = """
    MyMarkdownWindow {
        width: 50%;
        height: 50%;
        border: round;
    }
    """

    BINDINGS = [
        ("space", "quit", "Quit"),
    ]

    def __init__(self, markdown: str | None = None, title: str | None = None, *args, **kwargs):
        self.markdown = markdown
        self.title = title
        self.status = True
        super().__init__(*args, **kwargs)

    def on_mount(self) -> None:
        if self.title is not None:
            self.border_title = self.title
        self.border_subtitle = "Press 'space' to close this previewer"
    
    def compose(self) -> ComposeResult:
        yield Markdown(self.markdown)

    def action_quit(self) -> None:
        self.status = False

    @classmethod
    async def my_markdown_window(cls, parent: Widget, markdown: str, title: str | None = None, *args, **kwargs) -> None:
        markdown_window_instance = MyMarkdownWindow(markdown=markdown, title=title, classes = "top_layer", *args, **kwargs)
        await parent.mount(markdown_window_instance)
        markdown_window_instance.focus()
        while markdown_window_instance.status:
            await asyncio.sleep(0.1)
        markdown_window_instance.remove()


class _SelectPath(Widget):
    # Not used, switched to using SelectPathDialog by Screen. Kept as future reference.

    DEFAULT_CSS = """
    SelectPath {
        width: 40%;
        height: 50%;
        border: double;
    }
    #bottom {
        height: 3;
        align: center bottom;
    }
    Input {
        height: 1;
        border: solid gray;
    }
    #cancel {
        offset-x: -15%;
    }
    #confirm {
        offset-x: 15%;
    }
    """

    def __init__(self, label: str | Text = "", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.label = label
        self.status = True
        self.value = None

    def compose(self) -> ComposeResult:
        yield Label(self.label)
        yield Input(id="path")
        yield DirectoryTree("./user0", id="directory_tree")
        with Horizontal(id="bottom"):
            yield Button("Cancel", id="cancel")
            yield Button("Confirm", id="confirm")

    def on_button_pressed(self, event) -> None:
        if event.button.id == "confirm":
            self.value = self.query_one("#path").value
            self.status = False
        elif event.button.id == "cancel":
            self.value = None
            self.status = False

    async def on_input_submitted(self, message) -> None:
        self.value = message.value
        self.status = False

    def on_directory_tree_file_selected(self, event) -> None:
        self.query_one("#path").value = str(os.path.dirname(event.path)) + "/"

    @classmethod
    async def select_path(
        cls,
        parent: Widget,
        callback: Callable[[str], None] | Callable[[str], Awaitable[None]],
        label: str | Text = "",
        *args,
        **kwargs
    ) -> None:
        _instance = _SelectPath(label=label, classes="top_layer", *args, **kwargs)
        await parent.mount(_instance)
        while _instance.status:
            await asyncio.sleep(0.1)
        value = _instance.value
        _instance.remove()
        if value is not None:
            if inspect.iscoroutinefunction(callback):
                await callback(value)
            elif callable(callback):
                callback(value)
            else:
                raise ValueError("Callback must be callable or a coroutine function")


class AppStart(Widget):
    
    DEFAULT_CSS = """
    AppStart {
        border: none;
    }
    AppStart RichLog {
        overflow: auto auto;
        scrollbar-size: 1 1;
        background: black;
    }
    AppStart Static {
        height: 14;
        text-align: center;
        background: black;
    }
    """

    def __init__(self, my_app, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.my_app = my_app

    def compose(self) -> ComposeResult:
        self.static = Static()
        yield self.static
        self.rich_log = RichLog(highlight=True, markup=True)
        yield self.rich_log
    
    def on_show(self):
        width = self.static.content_size.width
        if width >= 50:
            text = Text(
                textwrap.dedent(
                    """
                      $$$$$$\\  $$$$$$$\\ $$$$$$$$\\ $$\\   $$\\ $$$$$$\\ 
                      $$  __$$\\ $$  __$$ \\__$$  __|$$ |  $$ |\\_$$  _|
                    $$ /  \\__|$$ |  $$ |  $$ |   $$ |  $$ |  $$ |  
                    $$ |$$$$\\ $$$$$$$  |  $$ |   $$ |  $$ |  $$ |  
                    $$ |\\_$$ |$$  ____/   $$ |   $$ |  $$ |  $$ |  
                    $$ |  $$ |$$ |        $$ |   $$ |  $$ |  $$ |  
                     \\$$$$$$  |$$ |        $$ |   \\$$$$$$  |$$$$$$\\ 
                       \\______/ \\__|        \\__|    \\______/ \\______|
                    """
                )
            )
            text += Text(f"\nVersion: {self.my_app.app_version}\n")
            text += Text("Copyright (c) 2023 happyapplehorse\n")
            text += Text("e-mail: chaoxueao@gmial.com\n")
            text += Text("Note: Voices in this app are AI-generated, not human-spoken.\n")
            text += "-" * width
            self.static.update(text)
        else:
            self.static.remove()
    
    def get_rich_log(self):
        return self.rich_log

    def set_init_func(self, func: Coroutine):
        self.init_func = func

    async def app_init(self) -> bool:
        status = await self.init_func
        if status is False:
            return False
        self.static.update("")
        self.rich_log.clear()
        height = self.my_app.size.height
        self.animate("offset", value=Offset(0, height), duration=0.7, easing="linear")
        await asyncio.sleep(0.7)
        self.remove()
        return True


class NoPaddingButton(Button):
    def render(self) -> TextType:
        label = Text.assemble(self.label)
        label.stylize(self.text_style)
        return label



import openai
from openai import OpenAI, AsyncOpenAI

from .openai_settings_from_dot_env import openai_settings_from_dot_env


OpenAIClient = OpenAI | AsyncOpenAI


def openai_api(dot_env_path: str | None):
    assert dot_env_path, "'dot_env_path' can not be None or empty."
    openai_key, org_id = openai_settings_from_dot_env(dot_env_path)
    openai.api_key = openai_key
    return openai

def openai_api_client(dot_env_path: str | None, async_client: bool = False, **kwargs) -> OpenAIClient:
    assert dot_env_path, "'dot_env_path' can not be None or empty."
    openai_key, org_id = openai_settings_from_dot_env(dot_env_path)
    if async_client is True:
        client = AsyncOpenAI(api_key=openai_key, organization=org_id, **kwargs)
    else:
        client = OpenAI(api_key=openai_key, organization=org_id, **kwargs)
    return client



class ManagerError(Exception):
    ...


class HandlerNotRegisterError(ManagerError):
    def __init__(self, handler, manager=None):
        self.handler = handler
        self.manager = manager
    
    def __str__(self):
        if self.manager is None:
            return f"Handler: {self.handler} is not registered in manager."
        else:
            return f"Handler: {self.handler} is not registered in manager: {self.manager}."


class JobNotRegisterError(ManagerError):
    def __init__(self, job, manager=None):
        self.job = job
        self.manager = manager
    
    def __str__(self):
        if self.manager is None:
            return f"Job: {self.job} is not registered in manager."
        else:
            return f"Job: {self.job} is not registered in manager: {self.manager}."



import copy
import logging
from typing import Iterable

from openai import OpenAI

from .context import OpenaiContext
from .openai_error import OpenaiErrorHandler
from .openai_tokens_truncate import trim_excess_tokens
from .utils.openai_api import OpenAIClient
from .utils.tokens_num import tokens_num_for_functions_call


gptui_logger = logging.getLogger("gptui_logger")


def chat_service_for_inner(
        messages_list: list, 
        context: OpenaiContext,
        openai_api_client: OpenAIClient,
        **kwargs,
    ) -> Iterable:
    
    inner_context = copy.deepcopy(context)
    
    for one_message in messages_list:
        inner_context.chat_context_append(message=one_message)
    
    # update parameters
    parameters = inner_context.parameters
    parameters.update({"stream": True})
    parameters.update(**kwargs)

    if tools_info := parameters.get("tools"):
        offset_tokens_num = -tokens_num_for_functions_call(tools_info, model=inner_context.parameters["model"])
    else:
        offset_tokens_num = 0
    trimmed_messages = trim_excess_tokens(inner_context, offset=offset_tokens_num)
    
    # Delete the tool reply messages at the beginning of the information list.
    # This is because if the information starts with a function reply message,
    # it indicates that the function call information has already been truncated.
    # The OpenAI API requires that function reply messages must be responses to function calls.
    # Therefore, if the function reply messages are not removed, it will result in an OpenAI API error.
    while trimmed_messages and trimmed_messages[0].get("role") == "tool":
        trimmed_messages.pop(0)

    try:
        response = openai_api_client.with_options(timeout=20.0).chat.completions.create(
            messages=trimmed_messages,
            **parameters,
            )
    except Exception as e:
        gptui_logger.debug('----trimmed_messages----in chat inner')
        gptui_logger.debug(trimmed_messages)
        # The OpenAI API interface is a time-consuming synchronous interface, so it should be called in a new thread, hence there is no event loop here.
        OpenaiErrorHandler().openai_error_handle(error=e, context=inner_context, event_loop=False)
        raise e
    return response



from textual.message import Message


class CommonMessage(Message):
    def __init__(self, message_name: str, message_content) -> None:
        self.message_name = message_name
        self.message_content = message_content
        super().__init__()



import uuid
import logging
from logging import Logger

from qdrant_client import QdrantClient
from semantic_kernel.connectors.memory.qdrant.qdrant_memory_store import QdrantMemoryStore
from semantic_kernel.memory.memory_record import MemoryRecord
from semantic_kernel.utils.null_logger import NullLogger
from qdrant_client import models as qdrant_models


gptui_logger = logging.getLogger("gptui_logger")


class QdrantVector(QdrantMemoryStore):
    
    def __init__(
        self,
        vector_size: int,
        url: str | None = None,
        port: int | None = 6333,
        logger: Logger | None = None,
        local: bool | None = False,
    ) -> None:
        """Initializes a new instance of the QdrantMemoryStore class.

        Arguments:
            logger {Optional[Logger]} -- The logger to use. (default: {None})
        """
        if local:
            if url:
                try:
                    self._qdrantclient = QdrantClient(path=url)
                except KeyError as e:
                    gptui_logger.error(
                        f"An error occurred while initializing the local vector database. Database path: {url}. Error: {repr(e)} "
                        "Warning: Rebuilding of the vector database may be required."
                    )
            else:
                self._qdrantclient = QdrantClient(location=":memory:")
        else:
            self._qdrantclient = QdrantClient(url=url, port=port)

        self._logger = logger or NullLogger()
        self._default_vector_size = vector_size

    async def _convert_from_memory_record_async(
        self, collection_name: str, record: MemoryRecord
    ) -> qdrant_models.PointStruct:
        if record._key is not None and record._key != "":
            pointId = record._key

        else:
            existing_record = await self._get_existing_record_by_payload_id_async(
                collection_name=collection_name,
                payload_id=record._id,
            )

            if existing_record:
                pointId = str(existing_record.id)
            else:
                pointId = str(uuid.uuid4())

        payload = record.__dict__.copy()
        payload["storage_status"] = "unsaved"
        embedding = payload.pop("_embedding")

        return qdrant_models.PointStruct(
            id=pointId, vector=embedding.tolist(), payload=payload
        )

    async def collection_save(self, collection_name: str) -> qdrant_models.UpdateResult:
        filter = qdrant_models.Filter(
            must=[
                qdrant_models.FieldCondition(
                    key="storage_status",
                    match=qdrant_models.MatchAny(any=["unsaved", "cached"]),
                )
            ]
        )

        update_result = self._qdrantclient.set_payload(
            collection_name=str(collection_name),
            payload={"storage_status": "saved"},
            points=filter,
        )

        return update_result

    async def collection_cache(self, collection_name: str) -> qdrant_models.UpdateResult:
        filter = qdrant_models.Filter(
            must=[
                qdrant_models.FieldCondition(
                    key="storage_status",
                    match=qdrant_models.MatchValue(value="unsaved"),
                )
            ]
        )
        
        update_result = self._qdrantclient.set_payload(
            collection_name=str(collection_name),
            payload={"storage_status": "cached"},
            points=filter,
        )

        return update_result
    
    async def collection_clean(self, collection_name: str) -> qdrant_models.UpdateResult:
        filter = qdrant_models.Filter(
            must=[
                qdrant_models.FieldCondition(
                    key="storage_status",
                    match=qdrant_models.MatchAny(any=["unsaved", "cached"]),
                )
            ]
        )

        update_result = self._qdrantclient.delete(
            collection_name=collection_name,
            points_selector=qdrant_models.FilterSelector(filter=filter),
        )
        
        return update_result

    async def collection_count(self, collection_name: str) -> tuple[int, int, int]:
        def make_filter(storage_status: str):
            filter = qdrant_models.Filter(
                must=[
                    qdrant_models.FieldCondition(
                        key="storage_status",
                        match=qdrant_models.MatchValue(value=storage_status),
                    )
                ]
            )
            return filter

        count_saved = self._qdrantclient.count(
            collection_name=collection_name,
            count_filter=make_filter("saved"),
            exact=True,
        )

        count_cached = self._qdrantclient.count(
            collection_name=collection_name,
            count_filter=make_filter("cached"),
            exact=True,
        )

        count_unsaved = self._qdrantclient.count(
            collection_name=collection_name,
            count_filter=make_filter("unsaved"),
            exact=True,
        )

        return count_saved.count, count_cached.count, count_unsaved.count



import importlib
import logging
import os
from abc import ABCMeta, abstractmethod

import semantic_kernel as sk
from semantic_kernel.template_engine.prompt_template_engine import PromptTemplateEngine

from .doc import Doc
from ..gptui_kernel.manager import Manager


gptui_logger = logging.getLogger("gptui_logger")


class UploadFileInterface(metaclass=ABCMeta):
    @abstractmethod
    async def import_file_to_context(self, *docs: Doc, input: str) -> str:
        ...


class Skills:
    def __init__(self, manager: Manager):
        self.manager = manager

    async def conversation_remember(self, conversation: str):
        to_string_function = self.manager.gk_kernel.sk_kernel.skills.get_function("conversation_service", "conversation_to_string")
        conversation_string = to_string_function(conversation)
        remember_function = self.manager.gk_kernel.sk_kernel.skills.get_function("conversation_service", "conversation_remember")
        summary = await remember_function.invoke_async(conversation_string)
        return str(summary)


class UploadFile(UploadFileInterface):
    
    UPLOAD_FILE_PROMPT_PATH = os.path.join(os.path.dirname(__file__), "gptui_basic_services", "templates", "upload_file_prompt.txt")
    
    def __init__(self):
        self.prompt_template_engine = PromptTemplateEngine()
        self.kernel = sk.Kernel()
        try:
            with open(UploadFile.UPLOAD_FILE_PROMPT_PATH, "r") as import_prompt:
                self.template_text = import_prompt.read()
        except FileNotFoundError:
            gptui_logger.error("File not found")
            #self.app.query_one("#status_region").update(Text("File not found",'yellow'))
        except IsADirectoryError:
            gptui_logger.error("Specified path is a directory, not a file")
            #self.app.query_one("#status_region").update(Text("Specified path is a directory, not a file",'yellow'))
        except UnicodeDecodeError:
            gptui_logger.error("File is not encoded properly")
            #self.app.query_one("#status_region").update(Text("File is not encoded properly",'yellow'))
        except IOError as e:
            #self.app.query_one("#status_region").update(Text(f"An I/O error occurred: {e}",'yellow'))
            gptui_logger.error(f"An I/O error occurred: {e}")
        except Exception as e:
            #self.app.query_one("#status_region").update(Text('Have not sucessfully read memory.','yellow'))
            gptui_logger.error("Read file failed. An unexpected error occurred: {e}")

    async def import_file_to_context(self, *docs: Doc, input: str) -> str:
        if len(docs) >= 1:
            files_content = ''
            for index, doc in enumerate(docs[:-1]):
                file_title = f"===== Document #{index + 1} {doc.name + doc.ext} =====\n\n"
                files_content += file_title
                files_content += doc.content
                files_content += "\n\n" + "=" * (len(file_title) - 2) + "\n\n"
            file_title_last = f"===== Document #{len(docs)} {docs[-1].name + docs[-1].ext} =====\n\n"
            files_content += file_title_last
            files_content += docs[-1].content
            files_content += "\n\n" + "=" * (len(file_title_last) - 2)
        else:
            raise ValueError("There is no document!")
        context = self.kernel.create_new_context()
        context["input"] = input
        context["file_content"] = files_content
        result_prompt = await self.prompt_template_engine.render_async(template_text=self.template_text, context=context)
        return result_prompt



import itertools
import json
import logging

import chardet
import httpx
from lxml import html
from semantic_kernel.orchestration.sk_context import SKContext
from semantic_kernel.skill_definition import sk_function, sk_function_context_parameter

from gptui.gptui_kernel.manager import auto_init_params
from gptui.models.utils.config_from_dot_env import config_from_dot_env


gptui_logger = logging.getLogger("gptui_logger")


class WebServe:
    def __init__(self, manager):
        self.manager = manager

    @auto_init_params("0")
    @classmethod
    def get_init_params(cls, manager) -> tuple:
        return (manager,)

    @sk_function(
        description="Searching for information on the Internet through Google",
        name="google_search",
    )
    @sk_function_context_parameter(
        name="query",
        description="The keywords to search",
    )
    @sk_function_context_parameter(
        name="num",
        description="Number of search results to return. Valid values are integers between 1 and 10, inclusive",
    )
    @sk_function_context_parameter(
        name="start",
        description=("The index of the first result to return. The default number of results per page is 10, "
                     "so &start=11 would start at the top of the second page of results. Notice: start + num have to be less than 100."),
        default_value="1",
    )
    def google_search(self, context: SKContext) -> str:
        """
        {"status_codes":status_codes, "query": query, "reslut_list":[{result}]}
        status_codes:
            901: internet react error
            902: search empty
            903: input parameters error
        """
        config = config_from_dot_env(self.manager.dot_env_config_path)
        google_key = config.get("GOOGLE_KEY")
        cx = config.get("GOOGLE_CX")
        assert google_key is not None and cx is not None
        url = 'https://www.googleapis.com/customsearch/v1'
        
        query = context["query"]
        num = int(context["num"])
        start = int(context["start"])
        result_metadata = {"status_codes":None, "query":query, "result_list":[]}
        
        if not query:
            result_metadata["status_codes"] = 902
            return json.dumps(result_metadata, ensure_ascii=False)
        if (not 1 <= num <= 10) or start < 0 or num + start > 100:
            result_metadata["status_codes"] = 903
            return json.dumps(result_metadata, ensure_ascii=False)

        parameters = {
                "key": google_key,
                "cx": cx,
                "q": query,
                "num": num,
                "start": start,
                }
        try:
            with httpx.Client() as client:
                response = client.request("GET", url = url, params = parameters)
            results = response.json()
        except:
            result_metadata["status_codes"] = 901
            return json.dumps(result_metadata, ensure_ascii=False)
        else:
            result_metadata["status_codes"] = response.status_code
            if response.status_code == 200:
                for item in results['items']:
                    result_metadata["result_list"].append({"title":item["title"], "url":item["link"], "snippet":item["snippet"]})
            
            # result = json.dumps(result_metadata, ensure_ascii=False)
            # to format
            result_list = result_metadata["result_list"]
            content = f'Folllwing are the search results of "{query}":\n'
            for i, result in enumerate(result_list):
                content += f'{i+1}.\n'
                content += 'Title: ' + result["title"] + '\n'
                content += 'Snippet: ' + result["snippet"] + '\n'
                content += 'URL: ' + result["url"] + '\n\n'
            return content

    @sk_function(
        description="Get the text content of given webpage's url",
        name="web_parse",
    )
    @sk_function_context_parameter(
        name="url",
        description="The url of webpage",
    )
    @sk_function_context_parameter(
        name="table",
        description="True or False, indicate whether a table in html should be parsed",
        default_value="False"
    )
    def web_parse(self, context: SKContext) -> str:
        """Extract the title, if present.
        Extract all p paragraphs, various levels of subheadings, ordered lists, unordered lists, tables, and horizontal rules.
        """
        
        url = context["url"]
        ul = True
        ol = True
        dl = True
        table = bool(context["table"])
        title = True
        start_with_h = True
        a_href = True
        br = True
        blockquote = True
        img = False

        etree = html.etree
        
        def is_descendant(child, ancestor_check: list) -> bool:
            ancestor_list = child.xpath("./ancestor::*")
            for ancestor in ancestor_check:
                if ancestor in ancestor_list:
                    return True
            return False

        def p_get(p) -> str:
            text = '\n'
            all_node = p.xpath("./node()")
            i = 0
            for content in all_node:
                i += 1
                if isinstance(content, str):
                    if content.strip():
                        text += content.strip()
                elif content in node_get(p):
                    txt = content_get_iterable_distribute(content)
                    if i == 1:
                        if txt.startswith('\n'):
                            text += txt[1:]
                        else:
                            text += txt
                else:
                    text += ''.join(content.xpath(".//text()"))
            return text
        def ul_get(ul) -> str:
            text = '\n'
            item_list = ul.xpath("./li")
            for item in item_list:
                text += u'\u2022   '
                i = 0
                for content in node_get(item):
                    i += 1
                    if isinstance(content, str):
                        if content.strip():
                            text += content.strip()
                    else:
                        txt = content_get_iterable_distribute(content)
                        if i == 1:
                            if txt.startswith('\n'):
                                text += txt[1:]
                            else:
                                text += txt
                        else:
                            if txt.startswith('\n'):
                                text += '\n    ' + txt[1:]
                            else:
                                text += txt
                text += '\n'
            return text[:-1]
        def ol_get(ol) -> str:
            text = '\n'
            item_list = ol.xpath("./li")
            num = 1
            for item in item_list:
                text += f'{num}.  '
                i = 0
                for content in node_get(item):
                    i += 1
                    if isinstance(content, str):
                        if content.strip():
                            text += content.strip()
                    else:
                        txt = content_get_iterable_distribute(content)
                        if i == 1:
                            if txt.startswith('\n'):
                                text += txt[1:]
                            else:
                                text += txt
                        else:
                            if txt.startswith('\n'):
                                text += '\n    ' + txt[1:]
                            else:
                                text += txt
                num += 1
                text += '\n'
            return text[:-1]
        def dl_get(dl) -> str:
            text = '\n'
            item_list = dl.xpath("./*[name() = 'dt' or name() = 'dd']")
            for item in item_list:
                if item.tag == 'dt':
                    i = 0
                    for content in node_get(item):
                        i += 1
                        if isinstance(content, str):
                            if content.strip():
                                text += content.strip()
                        else:
                            txt = content_get_iterable_distribute(content)
                            if i == 1:
                                if txt.startswith('\n'):
                                    text += txt[1:]
                                else:
                                    text += txt
                            else:
                                text += txt
                    text += '\n'
                elif item.tag == 'dd':
                    i = 0
                    for content in node_get(item):
                        i += 1
                        if isinstance(content, str):
                            if content.strip():
                                text += '    ' + content.strip()
                        else:
                            txt = content_get_iterable_distribute(content)
                            if i == 1:
                                if txt.startswith('\n'):
                                    text += '    ' + txt[1:]
                                else:
                                    text += '    ' + txt
                            else:
                                if txt.startswith('\n'):
                                    text += '\n    ' + txt[1:]
                                else:
                                    text += txt
                    text += '\n'
            return text[:-1]
        def table_get(table) -> str:
            text = '\n(Following is a table in MarkDown type:)\n'
            caption = table.xpath(".//caption")
            if caption:
                text += 'Caption:  ' + ''.join(caption[0].xpath("./text()")) + '\n'
            th_list = table.xpath(".//tr/th")
            text += '|'
            for item in th_list:
                text += ' ' + ''.join(item.xpath("./text()")) + ' |'
            text += '\n'
            text += '|---' * len(th_list) + '|\n'
            item_list = table.xpath(".//tr/td/..")
            for item in item_list:
                text += '|'
                for content in item.xpath("./td"):
                    for cont in node_get(content):
                        if isinstance(cont, str):
                            if cont.strip():
                                text += ' ' + cont.strip()
                        else:
                            txt = content_get_iterable_distribute(content)
                            if txt.startswith('\n'):
                                text += ' ' + txt[1:]
                            else:
                                text += ' ' + txt
                    text += ' |'
                text += '\n'
            return text[:-1]
        def h_get(h) -> str:
            text = '\n'
            i = 0
            for content in node_get(h):
                i += 1
                if isinstance(content, str):
                    if content.strip():
                        text += content.strip()
                else:
                    txt = content_get_iterable_distribute(content)
                    if i == 1:
                        if txt.startswith('\n'):
                            text += txt[1:]
                        else:
                            text += txt
                    else:
                        text += txt
            return text
        def a_get(a) -> str:
            text = ''
            for content in node_get(a):
                if isinstance(content, str):
                    if content.strip():
                        text += content.strip()
                else:
                    txt = content_get_iterable_distribute(content)
                    if txt.startswith('\n'):
                        text += txt[1:]
                    else:
                        text += txt
            text += f'{a.xpath("@href")}'
            return text
        def img_get(img) -> str:
            text = '\n[This is a image:  '
            if img.xpath("./@title"):
                text += 'title = ' + ''.join(img.xpath("./@title")) + '  '
            text += '{url = ' + ''.join(img.xpath("./@src")) + '}]'
            return text
        def blockquote_get(blockquote) -> str:
            text = '\n'
            i = 0
            for content in node_get(blockquote):
                i += 1
                if isinstance(content, str):
                    if content.strip():
                        text += content.strip()
                else:
                    txt = content_get_iterable_distribute(content)
                    if i == 1:
                        if txt.startswith('\n'):
                            text += txt[1:]
                        else:
                            text += txt
                    else:
                        text += txt
            return text
        def span_get(span) -> str:
            return '' + ''.join(span.xpath(".//text()"))
        
        def node_get(node):
            "include the text in the node itself by name()=''"
            content_list = node.xpath("./node()[name() = '' or name() = 'p' or name() = 'ul' or name() = 'ol' or name() = 'dl' or name() = 'blockquote' \
                    or name() = 'table' or name() = 'title' or starts-with(name(),'h') or (name() = 'a' and @href) or name() = 'img' or name() = 'br' or name() = 'span']")
            return content_list
        
        def content_get_iterable_distribute(content) -> str:
            if content.tag == "title":
                return h_get(content)
            elif content.tag in ['h1','h2','h3','h4','h5','h6']:
                return h_get(content)
            elif content.tag == "p":
                return p_get(content)
            elif content.tag == "ul":
                return ul_get(content)
            elif content.tag == "ol":
                return ol_get(content)
            elif content.tag == "dl":
                return dl_get(content)
            elif content.tag == "table":
                return table_get(content)
            elif content.tag == "a":
                return a_get(content)
            elif content.tag == 'hr' or content.tag == 'br':
                return '\n'
            elif content.tag == 'img':
                return img_get(content)
            elif content.tag == 'blockquote':
                return blockquote_get(content)
            elif content.tag == 'span':
                return span_get(content)
            else:
                return ''

        headers = {
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/111.0.0.0 Safari/537.36 Edg/111.0.1661.54'
                }
        
        # auto detect the character set
        def autodetect(content):
            return chardet.detect(content).get("encoding")

        result = {"status_code":None, "url":url, "web_text":''}
        
        try:
            with httpx.Client(default_encoding = autodetect) as client:
                response = client.request("GET", url = url, headers = headers, follow_redirects = True)
            page_text = response.text
        except:
            result["status_code"] = 901
            return json.dumps(result, ensure_ascii=False)
        else:
            tree = etree.HTML(page_text)
            xpath_str = "//p/..//*[name() = 'p'"
            if ul:
                xpath_str += " or name() = 'ul'"
            if ol:
                xpath_str += " or name() = 'ol'"
            if dl:
                xpath_str += " or name() = 'dl'"
            if table:
                xpath_str += " or name() = 'table'"
            if title:
                xpath_str += " or name() = 'title'"
            if start_with_h:
                xpath_str += " or starts-with(name(),'h')"
            if a_href:
                xpath_str += " or (name() = 'a' and @href)"
            if br:
                xpath_str += " or name() = 'br'"
            if blockquote:
                xpath_str += " or name() = 'blockquote'"
            if img:
                xpath_str += " or name() = 'img'"
            xpath_str += "]"
            content_list = tree.xpath(xpath_str)
            #content_list = tree.xpath("//p/..//*[name() = 'p' or name() = 'ul' or name() = 'ol' or name() = 'dl' or name() = 'table' \
            #    or name() = 'title' or starts-with(name(),'h') or (name() = 'a' and @href) or name() = 'br' or name() = 'blockquote' or name() = 'img']"
            #)
            text_list = []
            already_traversed_list = []
            for content in content_list:
                if is_descendant(content, already_traversed_list):
                    continue
                text_list.append(content_get_iterable_distribute(content))
                already_traversed_list.append(content)
            result["status_code"] = response.status_code
            result["web_text"] = ''.join(list(itertools.chain(*text_list)))
            return json.dumps(result, ensure_ascii=False)



import asyncio


def sync_wrapper(func):
    """
    Wrap a sync receiver to a async receiver.

    Usage example:
        result = await signal.send_async("sender", _sync_wrapper=sync_wrapper, message="message")
    """
    async def inner(*args, **kwargs):
        func(*args, **kwargs)
    return inner

def async_wrapper_with_loop(func):
    """
    Wrap a coroutine function receiver to a sync receiver.
    Suitable for cases where signals are sent within an event loop.
    It is recommended to directly use: result = await signal.send_async("sender", _sync_wrapper=sync_wrapper, message="message")

    Return: Task
    
    Example of await task:
        # Retrieve the Task objects returned by coroutine receivers and wait for their completion
        result = signal.send(sender, _async_wrapper=async_wrapper_with_loop, message="message")
        signal_tasks = [item[1] for item in result if getattr(item[0], "_async_inner_", False)]
        await asyncio.gather(*signal_tasks)
    """

    def inner(*args, **kwargs):
        task = asyncio.create_task(func(*args, **kwargs))
        return task
    # Add a coroutine marker to facilitate the identification of its Task object and wait for its completion.
    inner._async_inner_ = True
    return inner

def async_wrapper_without_loop(func):
    """
    Wrap a coroutine function receiver to a sync receiver.
    Suitable for cases where signals are sent without an event loop.

    Return: return the value returned from coroutine.
    
    Usage example:
        result = signal.send("sender", _async_wrapper=async_wrapper_without_loop, message="message")
    """
    def inner(*args, **kwargs):
        return asyncio.run(func(*args, **kwargs))
    return inner



import logging
import os

from semantic_kernel.orchestration.sk_context import SKContext
from semantic_kernel.skill_definition import sk_function, sk_function_context_parameter

from gptui.gptui_kernel.manager import auto_init_params
from gptui.models.doc import Doc
from gptui.views.common_message import CommonMessage


gptui_logger = logging.getLogger("gptui_logger")


class WriteFile:
    def __init__(self, manager):
        self.manager = manager
    
    @auto_init_params("0")
    @classmethod
    def get_init_params(cls, manager) -> tuple:
        return (manager,)

    @sk_function(
        description="Write a file.",
        name="write_file",
    )
    @sk_function_context_parameter(
        name="file_name",
        description="The name of the file, including the extension.",
    )
    @sk_function_context_parameter(
        name="file_content",
        description="The content to be written into the file."
    )
    def write_file(self, context: SKContext) -> str:
        file_name = context["file_name"]
        file_content = context["file_content"]
        self.manager.client.common_resources["temp_files_from_tube"] = {file_name: file_content}
        name, ext = os.path.splitext(file_name)
        document = Doc(doc_name=name, doc_ext=ext, pointer=self.manager.client.common_resources["temp_files_from_tube"][file_name], description="")
        # Only main thread can handle UI event correctly
        self.manager.client.post_message(CommonMessage(message_name="write_file", message_content=document))
        return "Write file sucessfully!" 



def safe_next(gen):
    """Avoiding conflicts between StopIteration of generators and StopIteration of coroutine functions in eventloop."""
    try:
        return ("OK", next(gen))
    except StopIteration as e:
        return ("DONE", e.value)

def safe_send(gen, value):
    """Avoiding conflicts between StopIteration of generators and StopIteration of coroutine functions in eventloop."""
    try:
        return ("OK", gen.send(value))
    except StopIteration as e:
        return ("DONE", e.value)



import textwrap
from blinker import signal


response_to_user_message_stream_signal = signal("response_to_user_message_stream",
    doc=textwrap.dedent(
        """
        Send the to_user part of response message, which usually should be displayed.
        position arg: sender
        kwargs:
            message {dict}:
                content: message from LLM to user in stream
                flag {Literal['content','end']}: 'content' means in stream, while 'end' means stream finished
        It should end with a signal to finish the stream as below:
            signal('response_to_user_message_stream').send(sender, message={'content':'', 'flag':'end'})
        """
    )
)
response_to_user_message_sentence_stream_signal = signal("response_to_user_message_sentence_stream",
    doc=textwrap.dedent(
        """
        Send the to_user part of response message in sentence, which is useful when outputting speech.
        position arg: sender
        kwargs:
            message {dict}:
                content: message from LLM to user in sentence stream
                flag {Literal['content','end']}: 'content' means in stream, while 'end' means stream finished
        It should end with a signal to finish the stream as below:
            signal('response_to_user_message_sentence_stream').send(sender, message={'content':'', 'flag':'end'})
        """
    )
)
notification_signal = signal("notification",
    doc=textwrap.dedent(
        """
        Sending notification-type information, such as status information, error messages, warning, etc.
        position arg: sender
        kwargs:
            message {dict}:
                content: notification
                flag {str}: type of notification, such as 'info', 'warning', 'error', etc.
        """
    )
)
response_auxiliary_message_signal = signal("response_auxiliary_message",
    doc=textwrap.dedent(
        """
        Sending auxiliary information, such as function call information, other internel messages, etc.
        position arg: sender
        kwargs:
            message {dict}:
                content: auxiliary message
                flag {str}: type of auxiliary message
        """
    )
)
chat_context_extend_signal = signal("chat_context_extend",
    doc=textwrap.dedent(
        """
        Sending a notification to save the chat context.
        position arg: sender
        kwargs:
            message {dict}:
                content {dict}: chat context information
                    {
                        "messages" {list}: messages
                        "context": context which is extended to
                    }
                flag:""
        """
    )
)
chat_context_extend_for_sending_signal = signal("chat_context_extend_for_sending",
    doc=textwrap.dedent(
        """
        Sending a notification to append or extend the chat context for sending.
        position arg: sender
        kwargs:
            message {dict}:
                content: dict, chat context information
                    {
                        "messages" {list}: messages
                        "context": context which is extended to
                    }
                flag:""
        """
    )
)
common_message_signal = signal("common_message",
    doc=textwrap.dedent(
        """
        Designed to send general messages.
        Different message structures can be achieved through flexible use of 'content' and 'flag'.
        position arg: sender
        kwargs:
            message {dict}:
                content: message content
                flag {str}: type of the common message
        """
    )
)



import logging
import random
import threading
from typing import Iterable, AsyncIterable

from agere.commander import PASS_WORD, Job, Callback, tasker

from .blinker_wrapper import sync_wrapper
from .context import Context
from .role import Role
from .signals import (
    common_message_signal,
    notification_signal,
    response_auxiliary_message_signal,
)
from ..gptui_kernel.manager import ManagerInterface


gptui_logger = logging.getLogger("gptui_logger")


class ResponseJob(Job):
    def __init__(
        self,
        manager: ManagerInterface,
        response: Iterable,
        context: Context,
        callback: Callback | None = None,
        at_receiving_start: list[dict] | None = None,
        at_receiving_end: list[dict] | None = None,
    ):
        super().__init__(callback=callback)
        self.response = response
        self.context = context
        self.manager = manager
        self.at_receiving_start = at_receiving_start
        self.at_receiving_end = at_receiving_end

    @tasker(PASS_WORD)
    async def task(self):
        ResponseHandler = self.manager.get_handler("ResponseHandler")
        handler = ResponseHandler(self.manager, self.context).handle_response(
            response=self.response,
            at_receiving_start=self.at_receiving_start,
            at_receiving_end=self.at_receiving_end,
        )
        return handler


class GroupTalkManager(Job):
    def __init__(self, manager: ManagerInterface, user_name: str = "admin"):
        self._speaking = None
        self._ai_care_rlock = threading.RLock()
        self.running = False
        self.manager = manager
        self.user_name = user_name
        self.group_talk_manager_id: int | None = None
        self.roles: dict[str, Role] = {}
        self.user_talk_buffer = []
        self.ai_care_message_buffer: list[dict] = []
        super().__init__()
    
    @property
    def speaking(self) -> str | None:
        return self._speaking

    @speaking.setter
    def speaking(self, value: str | None):
        with self._ai_care_rlock:
            if value is None:
                messages = [{"role": "user", "name": self.user_name, "content": message} for message in self.user_talk_buffer]
                if messages:
                    self.ai_care_message_buffer = []
                    self._speaking = self.user_name
                    common_message_signal.send(
                        self,
                        _async_wrapper=async_wrapper_with_loop,
                        message={
                            "content": messages,
                            "flag": "group_talk_user_message_send",
                        },
                    )
                    response_dict = {}
                    items = list(self.roles.items())
                    # Randomly shuffle the order to give each role an equal opportunity to speak.
                    random.shuffle(items)
                    for role_name, role in items:
                        response_dict[role_name] = role.chat(message=messages)
                    GroupTalkHandler = self.manager.get_handler("GroupTalkHandler")
                    # Because at any given moment only one role can speak, it is safe when multiple dialogue tasks are running in parallel;
                    # there will be no inconsistency in the role's chat history.
                    self.call_handler(GroupTalkHandler().handle_response(response_dict))
                    self.user_talk_buffer = []
                    for role in self.roles.values():
                        role.ai_care_update()
            else:
                for role in self.roles.values():
                    role.ai_care.cancel_current_task()
                self.ai_care_message_buffer = []
            self._speaking = value
    
    async def set_speaking_to_none(self):
        with self._ai_care_rlock:
            self.speaking = None
            dose_send = await self.send_ai_care_message()
            if not dose_send:
                for role in self.roles.values():
                    role.ai_care_update()

    async def send_ai_care_message(self) -> bool:
        message = self.ai_care_message_buffer[0] if self.ai_care_message_buffer else None
        subsequent_messages = self.ai_care_message_buffer[1:]
        if message:
            role_name = message["name"]
            self.speaking = role_name
            try:
                message_content = await self.ai_care_response(role_name=role_name, response=message["content"])
            except Exception as e:
                await self.set_speaking_to_none()
                self.ai_care_message_buffer.extend(subsequent_messages)
                gptui_logger.info(f"Encountered an error when receiving the AICare speech content of group chat member {role_name}, error: {e}")
            else:
                await self.put_job(TalkToAll(message_content=message_content, message_from=role_name))
                # self.ai_care_message_buffer = [] # This operation has already been performed when setting speaking=role_name
                return True
        return False

    def create_role(self, role: Role, role_name: str) -> bool:
        if role_name in self.roles:
            notification_signal.send(
                self,
                _async_wrapper=async_wrapper_with_loop,
                message={
                    "content": f"The name of '{role_name}' has already existed.",
                    "flag": "warning",
                }
            )
            return False
        self.roles[role_name] = role
        return True

    def close_group_talk(self):
        self.running = False
        
    @tasker(PASS_WORD)
    async def task(self):
        self.running = True
        GroupTalkHandler = self.manager.get_handler("GroupTalkHandler")
        return GroupTalkHandler().wait_for_termination(self)

    async def ai_care_response(self, role_name: str, response: AsyncIterable) -> str:
        chunk_list = []
        async for chunk in response:
            if not chunk:
                continue
            await response_auxiliary_message_signal.send_async(
                self,
                _sync_wrapper=sync_wrapper,
                message={
                    "content": {
                        "flag": "content",
                        "content": {
                            "role": "assistant",
                            "name": role_name,
                            "content": chunk,
                            "group_talk_manager_id": self.group_talk_manager_id,
                        },
                    },
                    "flag": "group_talk_response",
                },
            )
            chunk_list.append(chunk)
        full_response_content = ''.join(chunk_list)
        await response_auxiliary_message_signal.send_async(
            self,
            _sync_wrapper=sync_wrapper,
            message={
                "content": {
                    "flag": "end",
                    "content": {"role": "assistant", "name": role_name, "content": "", "group_talk_manager_id": self.group_talk_manager_id},
                },
                "flag": "group_talk_response",
            },
        )
        return full_response_content


class TalkToAll(Job):
    def __init__(self, message_content: str, message_from: str):
        self.message_content = message_content
        self.message_from = message_from
        super().__init__()

    @tasker(PASS_WORD)
    async def task(self):
        talk_manager = self.ancestor_chain[-2]
        assert isinstance(talk_manager, GroupTalkManager)
        if not talk_manager.running:
            return
        response_dict = {}
        all_roles = list(talk_manager.roles.items())
        # Randomly shuffle the order to give each role an equal opportunity to speak.
        random.shuffle(all_roles)
        try:
            for role_name, role in all_roles:
                if role_name == self.message_from:
                    response_dict[role_name] = role.chat(message={"role": "assistant", "content": self.message_content})
                else:
                    response_dict[role_name] = role.chat(message={"role": "user", "name": self.message_from, "content": self.message_content})
        except Exception as e:
            gptui_logger.info(f"Encountered an error when the group chat member {self.message_from} talks to all, error: {e}")
        finally:
            await talk_manager.set_speaking_to_none()
        GroupTalkHandler = talk_manager.manager.get_handler("GroupTalkHandler")
        return GroupTalkHandler().handle_response(response_dict)
